import sys
import time
import subprocess


def schedule_reboot(when="Sun 06:00:00"):
    """Create a timer+service to reboot the server at a regular interval,
    e.g. every sunday. The default value for when is "Sun 06:00:00".
    """
    timer = mypaas_reboot_timer.replace("WHEN", when)
    service = mypaas_reboot_service

    with open("/etc/systemd/system/mypaas_reboot.timer", "bw") as f:
        f.write(timer.encode())
    with open("/etc/systemd/system/mypaas_reboot.service", "bw") as f:
        f.write(service.encode())
    time.sleep(0.1)
    try:
        subprocess.check_call(["systemctl", "daemon-reload"])
        subprocess.check_call(["systemctl", "restart", "mypaas_reboot.timer"])
        subprocess.check_call(["systemctl", "enable", "mypaas_reboot.timer"])
    except subprocess.SubprocessError:
        sys.exit("Could not create mypaas reboot timer")


mypaas_reboot_timer = """

[Unit]
Description=MyPaas reboot timer

[Timer]
OnCalendar=WHEN

[Install]
WantedBy=timers.target
"""

mypaas_reboot_service = """
[Unit]
Description=MyPaas reboot service

[Service]
User=root
ExecStart=/bin/bash -c 'reboot now'
RestartSec=20
"""



"""
The mypaas.stats subpackage represents a service to collect and view
all sorts of stats of your PaaS. All services can push measurements
over UDP.

Dependencies: fastuaparser, asgineer, pscript,
"""

# flake8: noqa
from .monitor import Monitor
from .collector import StatsCollector
from .receiver import UdpStatsReceiver
from .server import stats_handler



"""
The logic to serve the stats dashboard.

MyPaas sitemap:

    /           -> dashboard home, showing server stats, and links
    /dashboard  -> the Traefik dashboard (hosted by Traefik)
    /stats      -> stat pages, select groups and range via query params
    /daemon     -> very basic daemon info page (hosted by mypaasd)

"""

import os
import json
import time
import platform

import psutil
import pscript
import asgineer

from .client_style import CSS


START_TIME = time.time()
THIS_DIR = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(THIS_DIR, "client_code.py"), "rb") as f:
    JS = pscript.py2js(f.read().decode())


static_assets = {"style.css": CSS, "client.js": JS}
asset_handler = asgineer.utils.make_asset_handler(static_assets)


async def stats_handler(request, collector):
    """The main http handler to serve stats data."""

    if request.method != "GET":
        return 405, {}, "invalid method"

    if request.path == "/":
        groups = collector.get_groups()
        # Allow sorting group while, honoring the numeric suffix
        group_keys = {}
        for group in groups:
            pre, _, post = group.rpartition(".")
            if post.isnumeric():
                group_keys[group] = pre + "." + post.rjust(5, "0")
        # Collect groups in super-groups
        groups_grouped = {}
        for group in groups:
            if group in ("system", "stats", "traefik", "daemon"):
                base_group = "MyPaas server"
            else:
                base_group = group.split(".")[0]
            groups_grouped.setdefault(base_group, []).append(group)
        # Produce links
        links = []
        for base_group, groups in groups_grouped.items():
            if base_group != "MyPaas server":
                groups.sort(key=lambda g: group_keys.get(g, g))
            # Link for base group?
            if len(groups) > 1:
                groups_str = ",".join(groups)
                link = f"<a href='/stats?groups={groups_str}'>{base_group}</a>"
                links.append("<li>" + link + "</li>")
                links.append("<ul>")
            # Link for each group
            for group in groups:
                link = f"<a href='/stats?groups={group}'>{group}</a>"
                link += f"&nbsp;&nbsp;&nbsp;&nbsp;<span id='{group}-cpu'></span>"
                link += f"&nbsp;&nbsp;&nbsp;&nbsp;<span id='{group}-mem'></span>"
                links.append("<li>" + link + "</li>")
            if len(groups) > 1:
                links.append("</ul>")

        html = MAIN_HTML_TEMPLATE
        html = html.replace("{LINKS}", "\n".join(links))
        html = html.replace("{INFO}", get_system_info())
        html = html.replace("{INFO-STREAM}", get_system_info_stream())
        return 200, {}, html

    if request.path == "/stats":
        groups = request.querydict.get("groups", "")
        groups = [group.strip() for group in groups.split(",") if group.strip()]
        ndays = request.querydict.get("ndays", "")
        daysago = request.querydict.get("daysago", "")
        if groups:
            return get_webpage(
                collector, ndays, daysago, groups, title="MyPaas Monitor"
            )
        else:
            return 302, {"Location": "/"}, b""

    elif request.path == "/quickstats":

        quickstats = {"system-uptime": _uptime()}

        # Add system measurements
        for name, group, key in [
            ("system-cpu", "system", "cpu|num|%"),
            ("system-mem", "system", "mem|num|iB"),
            ("system-disk", "system", "disk|num|iB"),
            ("system-connections", "traefik", "open connections|num"),
            ("system-rtime", "traefik", "duration|num|s"),
        ]:
            v = collector.get_latest_value(group, key)
            if v is not None:
                if key.endswith("|iB"):
                    v = (
                        f"{v/2**30:0.3f} GiB"
                        if "disk" in key
                        else f"{v/2**20:0.1f} MiB"
                    )
                elif key.endswith("|%"):
                    v = f"{v:0.1f} %"
                elif key.endswith("|s"):
                    v = f"{1000*v:0.1f} ms"
                else:
                    v = str(v)
            quickstats[name] = v

        # Add measurements for each group
        for group in collector.get_groups():
            quickstats[group + "-cpu"] = quickstats[group + "-mem"] = ""
            cpu = collector.get_latest_value(group, "cpu|num|%")
            mem = collector.get_latest_value(group, "mem|num|iB")
            if cpu is not None:
                quickstats[group + "-cpu"] = f"{cpu:0.1f} %"
                if mem is not None:
                    quickstats[group + "-mem"] = f"{mem/2**20:0.1f} MiB"

        return 200, {}, quickstats

    # elif request.path == "/statstream":
    #     # print("starting stat stream")
    #     raise NotImplementedError("Asigneer does not properly close streams.")
    #     # return 200, {"content-type": "text/plain"}, stat_streamer()

    else:
        fname = request.path.split("/")[-1]
        return await asset_handler(request, fname)


# async def stat_streamer():
#     # eek, asgineer does not seem to stop this when the connection is closed by the client
#     while True:
#         await asyncio.sleep(1.0)
#         t = asyncio.Task.current_task()
#         print("sending at ", time.time(), id(t))
#         yield f"hi {time.time()}"


def get_webpage(collector, ndays, daysago, groups, title=None, extra_info=None):
    """Generate a webpage with aggegation data from ndays1 ago to
    ndays2 ago (the order does not matter). Returns an complete html
    document as a string.

    Note that this call performs sync queries to a database, so you
    might want to asyncify the calling of this method.
    """
    # Query data, dump to json, and sanitize
    ndays, daysago = _normalize_ndays_and_daysago(ndays, daysago)
    # todo: we could asyncify this call. But only admins watch this page so ...
    data = json.dumps(collector.get_data(groups, ndays, daysago))
    data = data.replace("<", "&lt;").replace(">", "&gt;")
    info = {}  # adding info here will make it display as the first panel
    # Build page
    html = STATSVIEW_HTML_TEMPLATE.replace("{TITLE}", title or "Monitor")
    html = html.replace("{NDAYS}", str(ndays)).replace("{DAYSAGO}", str(daysago))
    html = html.replace("{INFO}", json.dumps(info))
    html = html.replace("{DATA_PER_DB}", data)
    return html


def get_system_info():
    info = {
        "server name": str(platform.node()),
        "server platform": str(platform.platform()),
        "server cpu-count": str(psutil.cpu_count()),
        "server mem": f"{psutil.virtual_memory().total / 2**20:0.0f} MiB",
        "server disk": f"{psutil.disk_usage('/').total / 2**30:0.1f} GiB",
    }
    html = "<table>"
    for key, value in info.items():
        html += f"<tr><td>{key}</td><td>{value}</td></tr>"
    html += "</table>"
    return html


def get_system_info_stream():
    info = {
        "": "&nbsp;" * 30,
        "system uptime": "<span id='system-uptime' />",
        "CPU usage": "<span id='_system-cpu' />",
        "mem usage": "<span id='_system-mem' />",
        "disk usage": "<span id='system-disk' />",
        "avg response time": "<span id='system-rtime' />",
        "open connections": "<span id='system-connections' />",
    }
    html = "<table>"
    for key, value in info.items():
        html += f"<tr><td>{key}</td><td>{value}</td></tr>"
    html += "</table>"
    return html


def _uptime():
    nsecs = time.time() - START_TIME
    if nsecs >= 3 * 86400:
        return f"{nsecs / 86400:0.1f} days"
    elif nsecs >= 3 * 3600:
        return f"{nsecs / 3600:0.1f} hours"
    elif nsecs >= 3 * 60:
        return f"{nsecs / 60:0.1f} minutes"
    else:
        return f"{nsecs:0.0f} seconds"


def _normalize_ndays_and_daysago(ndays, daysago):
    ndays = int(float(ndays)) if ndays else 3
    daysago = int(float(daysago)) if daysago else 0
    return max(1, ndays), max(0, daysago)


MAIN_HTML_TEMPLATE = """
<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8">
  <title>MyPaas dashboard</title>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="stylesheet" href="style.css">
</head>
<body>

<script>
var statgetter = function () {
    fetch('/quickstats')
        .then(function(response) {
            return response.json();
        }).then(function(data) {
            for (key in data) {
                var el = document.getElementById(key);
                if (el) { el.innerHTML = data[key]; }
                if (key.startsWith('system-')) {
                    el = document.getElementById("_" + key);
                    if (el) { el.innerHTML = data[key]; }
                }
            }
        });
    setTimeout(statgetter, 1000);
};
setTimeout(statgetter, 10);
</script>

<h1>MyPaas dashboard</h1>

<h2>System info</h2>
<div class="panelcontainer">
    <div class="panel"> <div class="content">
    {INFO}
    </div></div>
    <div class="panel"> <div class="content">
    {INFO-STREAM}
    </div></div>
</div>

<h2>MyPaas core services</h2>
<ul class='links'>
    <li><a href='/'>The stats server (this page)</a></li>
    <li><a href='/dashboard/'>The router (Traefik)</a></li>
    <li><a href='/daemon/'>The daemon (deploys)</a></li>
</ul>

<h2>Stats</h2>

<ul class='links'>
{LINKS}
</ul>

</body>
</html>
""".lstrip()


STATSVIEW_HTML_TEMPLATE = """
<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8">
  <title>{TITLE}</title>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="stylesheet" href="style.css">
</head>
<body>

<script>
var text_color = "#BBB";
var info = {INFO};
var data_per_db = {DATA_PER_DB};
var ndays = {NDAYS};
var daysago = {DAYSAGO};
</script>
<script src='client.js'></script>

<div id='topbar' class='topbar'>
<a href='/'><b>{TITLE}</b></a>&nbsp;&nbsp;
<div style='display:inline-block;'>
Showing
<a href="javascript:update_range('zoomin');"> ◀ </a>
{NDAYS}
<a href="javascript:update_range('zoomout');"> ▶ </a>
days
</div> <div style='display:inline-block;'>
until
<a href="javascript:update_range('newer');"> ◀ </a>
{DAYSAGO}
<a href="javascript:update_range('older');"> ▶ </a>
days ago.
</div>

&nbsp; &nbsp;
<a href="javascript:toggle_columns();">compact/wide panels</a>
<a href="javascript:toggle_utc();">utc/local time</a>
</div>

</body>
</html>
""".lstrip()



import os
import time
import datetime

from .monitor import Monitor, merge


class StatsCollector:
    """Central object that collects data, distributing it into different
    monitor objects (which are each backed by an sqlite db).
    """

    def __init__(self, db_dir):
        os.makedirs(db_dir, exist_ok=True)
        self._db_dir = db_dir
        self._monitors = {}
        self._available_groups = set()
        self._last_values = {}

        for fname in os.listdir(self._db_dir):
            if fname.endswith(".db"):
                self._available_groups.add(fname[:-3])

    def _get_db_name(self, group):
        return os.path.join(self._db_dir, group + ".db")

    def _get_monitor(self, group):
        try:
            return self._monitors[group]
        except KeyError:
            monitor = Monitor(self._get_db_name(group))
            self._monitors[group] = monitor
            self._available_groups.add(group)
            return monitor

    def put(self, group, stats):
        monitor = self._get_monitor(group)
        t = time.time()
        with monitor:
            for key, value in stats.items():
                self._last_values[group + ">" + key] = t, value
                monitor.put(key, value)

    def put_one(self, group, key, value):
        """Put a single value into the groups monitor, and return
        whether the value was accepted.
        """
        monitor = self._get_monitor(group)
        self._last_values[group + ">" + key] = time.time(), value
        with monitor:
            return monitor.put(key, value)

    def get_groups(self):
        """Get a tuple of groups known to this collector."""
        come_first = {"system", "stats", "traefik", "daemon"}
        come_last = {"other"}
        groups = self._available_groups.copy()
        groups1 = groups.intersection(come_first)
        groups3 = groups.intersection(come_last)
        groups2 = groups.difference(come_first | come_last)
        groups1 = tuple(sorted(groups1, key=lambda x: x.replace("sys", "_sys")))
        return groups1 + tuple(sorted(groups2)) + tuple(sorted(groups3))

    def get_latest_value(self, group, key):
        t, value = self._last_values.get(group + ">" + key, (0, None))
        etime = 5 if key == "cpu|num|%" else 60
        if time.time() - t < etime:
            return value
        else:
            return None

    def get_data(self, groups, ndays, daysago):
        """Get aggegation data from ndays ago to daysago. The
        result is a dict, in which the keys are the categores, and each
        value is a list of the aggregations in the corresponding
        group. The aggegations are combined (aggegregated further)
        if needed to keep the returned list to a reasonable size.

        Note that this call performs sync queries to a database, so you
        might want to asyncify the calling of this method.
        """
        assert isinstance(groups, list)

        # Get range of days to collect
        today = time.gmtime()  # UTC
        today = datetime.date(today.tm_year, today.tm_mon, today.tm_mday)
        one_day = datetime.timedelta(days=1)
        final_day = today - one_day * daysago
        first_day = today - one_day * (daysago + ndays - 1)

        t1 = int(time.mktime(first_day.timetuple()))
        t2 = int(time.mktime((final_day + one_day).timetuple()))

        # Collect all data
        data_per_group = {}

        for group in groups:

            # Get 10 min aggregations from monitor
            monitor = self._get_monitor(group)
            data = monitor.get_aggregations(first_day, final_day)

            # Determine level of aggregation: none, hour, day, month
            nchars = 20
            keys = {aggr["time_key"] for aggr in data}
            for n in [16, 15, 13, 10, 7]:  # min, 10 min, hour, day, month
                if len(keys) > 150:
                    nchars = n
                    keys = {key[:nchars] for key in keys}

            # Merge aggr's that have the same key.
            data2 = [{"time_key": "x"}]
            for aggr in data:
                key = aggr["time_key"][:nchars]
                if key == data2[-1]["time_key"]:
                    merge(data2[-1], aggr)
                else:
                    aggr = aggr.copy()
                    aggr["time_key"] = key
                    data2.append(aggr)

            # Put a stub aggregation at the beginning and end so that all figures
            # have the same time range.
            if len(data2) == 1:
                data = []
            else:
                data = data2
                x = {"time_key": "x"}
                x["time_start"] = x["time_stop"] = min(t1, data[1]["time_start"])
                data[0] = x.copy()
                x["time_start"] = x["time_stop"] = max(t2, data[-1]["time_stop"])
                data.append(x.copy())

            data_per_group[group] = data

        return data_per_group



"""
The code to deploy a service specified in a Dockerfile.
"""

import os
import time
import json
from urllib.parse import urlparse

from ..utils import dockercall
from ._auth import load_config


alphabet = "abcdefghijklmnopqrstuvwxyz"
identifier_chars = alphabet + alphabet.upper() + "0123456789" + "_"

# Cannot map a volume onto these
FORBIDDEN_DIRS = [
    "~/.ssh",
    "~/_mypaas",
]
for d in list(FORBIDDEN_DIRS):
    if d.startswith("~"):
        FORBIDDEN_DIRS.append(os.path.expanduser(d))


def clean_name(name, allowed_chars):
    """Make sure that the given name is clean,
    replacing invalid characters with a dash.
    """
    ok = identifier_chars + allowed_chars
    newname = "".join(c if c in ok else "-" for c in name)
    newname = newname.lstrip("-")
    if not newname:
        raise RuntimeError(f"No valid chars in name '{name}'.")
    return newname


def deploy(deploy_dir):
    """Deploy the current directory as a service. The directory must
    contain at least a Dockerfile. In most cases you should probably
    'mypaas push' from your work machine instead.
    """
    for step in get_deploy_generator(deploy_dir):
        print(step)


def get_deploy_generator(deploy_dir):
    """Get a generator that does the deploy, one step at a time, yielding
    a desciption of each step.
    """

    dockerfile = os.path.join(deploy_dir, "Dockerfile")

    # Get env-vars (e.g. secrets) from config
    config = load_config()
    secrets = config.get("env", {})

    # Init
    stripchars = "'\" \t\r\n"
    service_name = ""
    port = 80
    portmaps = []
    scale = None
    scale_option = "roll"
    urls = []
    volumes = []
    envvars = {}
    maxcpu = None
    maxmem = None
    healthcheck = None

    # Get configuration from dockerfile
    with open(dockerfile, "rt", encoding="utf-8") as f:
        for line in f.readlines():
            if not line.lstrip().startswith("#"):
                continue
            line = line.lstrip("# \t")
            if line.startswith("mypaas."):
                key, _, val = line.partition("=")
                key = key.strip(stripchars)
                val = val.strip(stripchars)

                if not val:
                    pass
                elif key == "mypaas.service":
                    service_name = val
                elif key == "mypaas.url":
                    url = urlparse(val)
                    if url.scheme not in ("http", "https") or not url.netloc:
                        raise ValueError("Invalid mypaas.url: {val}")
                    elif url.params or url.query or url.fragment:
                        raise ValueError("Too precise mypaas.url: {val}")
                    urls.append(url)
                elif key == "mypaas.volume":
                    volumes.append(val)
                elif key == "mypaas.port":
                    port = int(val)
                elif key == "mypaas.publish":
                    portmaps.append(val)
                elif key == "mypaas.scale":
                    for opt in ("safe", "roll"):
                        if opt in val:
                            scale_option = opt
                            val = val.replace(opt, "").strip()
                    scale = int(val)
                elif key == "mypaas.healthcheck":
                    parts = val.split()
                    if len(parts) != 3:
                        raise ValueError("Healthcheck must be /path interval timeout")
                    elif not parts[0].startswith("/"):
                        raise ValueError("Healthcheck path must start with '/'")
                    elif not parts[1].endswith(("ms", "s", "m" "h")):
                        raise ValueError(
                            "Healthcheck interval must be a durarion ending in 'ms', 's', 'm' or 'h'"
                        )
                    elif not parts[2].endswith(("ms", "s", "m" "h")):
                        raise ValueError(
                            "Healthcheck timeout must be a durarion ending in 'ms', 's', 'm' or 'h'"
                        )
                    healthcheck = {
                        "path": parts[0],
                        "interval": parts[1],
                        "timeout": parts[2],
                    }
                elif key == "mypaas.env":
                    val = val.strip()
                    if "=" in val:
                        k, _, v = val.partition("=")
                    elif val in secrets:
                        k, v = val, secrets[val]
                    else:
                        raise ValueError(
                            f"Env {val} is not found in ~/_mypaas/config.toml"
                        )
                    envvars[k.strip()] = v.strip()
                elif key == "mypaas.maxcpu":
                    maxcpu = str(float(val))
                elif key == "mypaas.maxmem":
                    assert all(c in "0123456789kmgtKMGT" for c in val)
                    maxmem = val
                else:
                    raise ValueError(f"Invalid mypaas deploy option: {key}")

    # We need at least an image name
    if not service_name:
        raise ValueError(
            "No service name given. Use '# mypaas.service=xxxx' in Dockerfile."
        )

    # Get clean names
    service_name = clean_name(service_name, ".-/")  # suited for an env var
    traefik_service_name = clean_name(service_name, "").rstrip("-") + "-service"
    traefik_service = f"traefik.http.services.{traefik_service_name}"

    # Collect info from all containers. Note that names can change but labels cannot.
    container_infos = get_containers_info(service_name)

    def label(x):
        cmd.append("--label=" + x)

    # Construct command to start the container
    cmd = ["run", "-d", "--restart=always"]

    # Apply limits
    if maxcpu:
        cmd.append(f"--cpus=" + maxcpu)
    if maxmem:
        cmd.append(f"--memory=" + maxmem)

    # Always use mypaas network, so services find each-other by container name.
    cmd.append(f"--network=mypaas-net")

    # Add portmappings to local system
    cmd.extend(["--publish=" + portmap for portmap in portmaps])

    if urls:
        label(f"traefik.enable=true")
        label(f"{traefik_service}.loadbalancer.server.port={port}")
        if healthcheck and scale and scale > 0:
            # Turning on the health check ensures that the load balancer won't use
            # the container until the server actually runs.
            label(
                f"{traefik_service}.loadbalancer.healthCheck.path={healthcheck['path']}"
            )
            label(
                f"{traefik_service}.loadbalancer.healthCheck.interval={healthcheck['interval']}"
            )
            label(
                f"{traefik_service}.loadbalancer.healthCheck.timeout={healthcheck['timeout']}"
            )
    for url in urls:
        router_name = clean_name(url.netloc + url.path, "").strip("-") + "-router"
        router_insec = router_name.rpartition("-")[0] + "-https-redirect"
        rule = f"Host(`{url.netloc}`)"
        if len(url.path) > 0:  # single slash is no path
            rule += f" && PathPrefix(`{url.path}`)"
        # Make sure that this rule is not also used in another service, otherwise
        # the URL will not work, and it may also disturb other rules for this service.
        for info in container_infos:
            if info["is_this_service"]:
                continue
            if any(rule == value for value in info["labels"].values()):
                raise ValueError(
                    f"URL {url.netloc + url.path} is already used in {info['name']}"
                )
        if url.scheme == "https":
            # Secure
            label(f"traefik.http.routers.{router_name}.rule={rule}")
            label(f"traefik.http.routers.{router_name}.entrypoints=web-secure")
            label(f"traefik.http.routers.{router_name}.tls.certresolver=default")
            label(f"traefik.http.routers.{router_name}.tls.options=intermediate@file")
            label(f"traefik.http.routers.{router_name}.middlewares=hsts-header@file")
            # Redirect
            label(f"traefik.http.routers.{router_insec}.rule={rule}")
            label(f"traefik.http.routers.{router_insec}.entrypoints=web")
            label(
                f"traefik.http.routers.{router_insec}.middlewares=https-redirect@file"
            )
        else:
            label(f"traefik.http.routers.{router_name}.rule={rule}")
            label(f"traefik.http.routers.{router_name}.entrypoints=web")
        # The stats server needs to be behind auth
        if service_name == "stats":
            label(f"traefik.http.routers.{router_name}.middlewares=auth@file")

    for volume in volumes:
        server_dir, _, container_dir = volume.partition(":")
        if server_dir.startswith("~"):
            server_dir = os.path.expanduser(server_dir)
        server_dir = os.path.realpath(server_dir)
        if not server_dir.startswith(os.path.expanduser("~")):
            raise ValueError(f"Cannot map a volume onto {server_dir}")
        elif any(server_dir.startswith(d) for d in FORBIDDEN_DIRS):
            raise ValueError(f"Cannot map a volume onto {server_dir}")
        os.makedirs(server_dir, exist_ok=True)
        cmd.append(f"--volume={server_dir}:{container_dir}")

    # Set user env vars
    for k, v in envvars.items():
        cmd.append(f"--env={k}={v}")

    # Set some env variables
    cmd.append(f"--env=MYPAAS_SERVICE={service_name}")
    cmd.append(f"--env=MYPAAS_SCALE={scale}")
    cmd.append(f"--env=MYPAAS_PORT={port}")
    # --env=MYPAAS_CONTAINER is set in the functions below.

    # Deploy!
    if scale and scale > 0:
        if scale_option == "roll":
            return _deploy_scale_roll(
                container_infos, deploy_dir, service_name, cmd, scale
            )
        else:
            return _deploy_scale_safe(
                container_infos, deploy_dir, service_name, cmd, scale
            )
    else:
        return _deploy_no_scale(container_infos, deploy_dir, service_name, cmd)


def _deploy_no_scale(container_infos, deploy_dir, service_name, prepared_cmd):
    image_name = clean_name(service_name, ".-:/")
    base_container_name = clean_name(image_name, ".-")
    new_name = f"{base_container_name}"

    yield ""
    yield f"deploying {service_name} to container {new_name}"
    time.sleep(1)

    yield "building image"
    dockercall("build", "--pull", "-t", image_name, deploy_dir)

    # There typically is one, but there may be more, if we had failed
    # deploys or if previously deployed with scale > 1
    old_ids = get_id_name_for_this_service(container_infos)
    unique = str(int(time.time()))

    yield f"renaming {len(old_ids)} container(s)"
    for i, id in enumerate(old_ids.keys()):
        try:
            dockercall("rename", id, base_container_name + f".old.{unique}.{i+1}")
        except Exception:
            yield "Rename failed. Probably a crashed container -> removing!"
            dockercall("rm", "-f", id, fail_ok=True)

    for id, name in old_ids.items():
        yield f"stopping container (was {name})"
        dockercall("stop", id, fail_ok=True)

    try:
        yield f"starting new container {new_name}"
        cmd = prepared_cmd.copy()
        cmd.append(f"--env=MYPAAS_CONTAINER={new_name}")
        cmd.extend([f"--name={new_name}", image_name])
        dockercall(*cmd)
    except Exception:
        yield "fail -> recovering"
        dockercall("rm", "-f", new_name, fail_ok=True)
        for id, name in old_ids.items():
            dockercall("rename", id, name, fail_ok=True)
            dockercall("start", id, fail_ok=True)
        raise
    else:
        yield f"removing {len(old_ids)} old container(s)"
        for id in old_ids.keys():
            dockercall("rm", id, fail_ok=True)

    yield "pruning"
    dockercall("container", "prune", "--force")
    dockercall("image", "prune", "--force")
    yield f"done deploying {service_name}"


def _deploy_scale_safe(container_infos, deploy_dir, service_name, prepared_cmd, scale):
    image_name = clean_name(service_name, ".-:/")
    base_container_name = clean_name(image_name, ".-")

    yield ""
    yield f"deploying {service_name} to containers {base_container_name}.1..{scale}"
    time.sleep(1)

    yield "building image"
    dockercall("build", "--pull", "-t", image_name, deploy_dir)

    old_ids = get_id_name_for_this_service(container_infos)
    unique = str(int(time.time()))

    yield f"renaming {len(old_ids)} current containers"
    for i, id in enumerate(old_ids.keys()):
        try:
            dockercall("rename", id, base_container_name + f".old.{unique}.{i+1}")
        except Exception:
            yield "Rename failed. Probably a crashed container -> removing!"
            dockercall("rm", "-f", id, fail_ok=True)

    for id, name in old_ids.items():
        yield f"stopping container (was {name})"
        dockercall("stop", id, fail_ok=True)

    # Keep track of started containers, in case we must shut them down
    new_pool = []

    try:
        for i in range(scale):
            new_name = f"{base_container_name}.{i+1}"
            yield f"starting new container {new_name}"
            new_pool.append(new_name)
            cmd = prepared_cmd.copy()
            cmd.append(f"--env=MYPAAS_CONTAINER={new_name}")
            cmd.extend([f"--name={new_name}", image_name])
            dockercall(*cmd)
    except Exception:
        yield "fail -> recovering"
        for name in new_pool:
            dockercall("stop", name, fail_ok=True)
            dockercall("rm", name, fail_ok=True)
        for id, name in old_ids.items():
            dockercall("rename", id, name, fail_ok=True)
            dockercall("start", id, fail_ok=True)
        raise
    else:
        yield f"removing {len(old_ids)} old containers"
        for id in old_ids.keys():
            dockercall("rm", id, fail_ok=True)

    yield "pruning"
    dockercall("container", "prune", "--force")
    dockercall("image", "prune", "--force")
    yield f"done deploying {service_name}"


def _deploy_scale_roll(container_infos, deploy_dir, service_name, prepared_cmd, scale):
    image_name = clean_name(service_name, ".-:/")
    base_container_name = clean_name(image_name, ".-")

    yield ""
    yield f"rolling deploy of {service_name} to containers {base_container_name}.1..{scale}"
    time.sleep(1)

    yield "building image"
    dockercall("build", "--pull", "-t", image_name, deploy_dir)

    old_ids = get_id_name_for_this_service(container_infos)
    unique = str(int(time.time()))

    yield f"renaming {len(old_ids)} current containers (and wait 2s)"
    for i, id in enumerate(old_ids.keys()):
        try:
            dockercall("rename", id, base_container_name + f".old.{unique}.{i+1}")
        except Exception:
            yield "Rename failed. Probably a crashed container -> removing!"
            dockercall("rm", "-f", id, fail_ok=True)

    # Give things a bit of time to settle
    time.sleep(2)

    # Prepare pools
    old_pool = list(old_ids.keys())  # we pop and stop containers from this pool
    new_pool = []  # we add started containers to this pool

    # Determine how long to wait each time before stopping an old container,
    # based on the assumption that a container boots within 5 seconds.
    # In essence, we dont want to close the last old container before the first
    # new container is fully operational.
    max_time_we_expect_a_container_to_boot = 5
    pause_per_step = 1 + max_time_we_expect_a_container_to_boot / max(1, len(old_pool))

    try:
        for i in range(scale):
            # Start up a new container
            new_name = f"{base_container_name}.{i+1}"
            yield f"starting new container {new_name} (and wait {pause_per_step:0.1f}s)"
            new_pool.append(new_name)
            cmd = prepared_cmd.copy()
            cmd.append(f"--env=MYPAAS_CONTAINER={new_name}")
            cmd.extend([f"--name={new_name}", image_name])
            dockercall(*cmd)
            # Stop a container from the pool
            if old_pool:
                time.sleep(pause_per_step)
                id = old_pool.pop(0)
                yield f"stopping old container (was {old_ids[id]})"
                dockercall("stop", id, fail_ok=True)
                time.sleep(0.5)  # Again give it time to settle
    except Exception:
        yield "fail -> recovering"
        for name in new_pool:
            dockercall("stop", name, fail_ok=True)
            dockercall("rm", name, fail_ok=True)
        for id, name in old_ids.items():
            dockercall("rename", id, name, fail_ok=True)
            if id not in old_pool:
                dockercall("start", id, fail_ok=True)
        raise
    else:
        for id in old_pool:
            yield f"stopping old container (was {old_ids[id]})"
            dockercall("stop", id, fail_ok=True)
        yield f"removing {len(old_ids)} old containers"
        for id in old_ids.keys():
            dockercall("rm", id, fail_ok=True)

    yield "pruning"
    dockercall("container", "prune", "--force")
    dockercall("image", "prune", "--force")
    yield f"done deploying {service_name}"


def get_id_name_for_this_service(container_infos):
    """Get a dict mapping id->name for all containers corresponding to the
    current service.
    """
    ids = []
    for info in container_infos:
        if info["is_this_service"]:
            ids.append((info["name"], info["id"]))  # name first, for sorting
    ids.sort()
    return {id: name for name, id in ids}


def get_containers_info(service_name):
    """Get a list of dicts with info on each currently running container."""
    # Get current container ids
    ids = dockercall("container", "ls", "--format", "{{.ID}}").split()
    # Get info for each container
    container_infos = []
    for id in ids:
        name_json = dockercall(
            "inspect", "--format", "{{.Name}}@{{json .Config.Labels}}", id
        )
        name, json_str = name_json.split("@", 1)
        container_infos.append(
            {"id": id, "name": name.lstrip("/"), "labels": json.loads(json_str)}
        )
    # Mark containers that match our service name
    base_container_name = clean_name(service_name, ".-")
    container_prefix = base_container_name + "."
    for info in container_infos:
        name = info["name"]
        is_this = name == base_container_name or name.startswith(container_prefix)
        info["is_this_service"] = is_this
    return container_infos



"""
The mypaas.daemon subpackage is started as a systemctl service, and
provides an API to push deploys and poll the PaaS status.

Dependencies: asgineer, psutil
"""

# flake8: noqa

from ._api import main_handler
from ._statsgen import SystemStatsProducer



import os
import io
import zipfile
from urllib.parse import quote

import requests

from ._keys import get_private_key
from ..utils import generate_uid


ignore_dirs = "__pycache__", "htmlcov", ".git", "node_modules"


def push(domain, dockerfile):
    """Push the given dockerfile (or directory containing a Dockerfile)
    to your PaaS, where it will be deployed as an app/service.
    """

    if domain.lower().startswith(("https://", "http://")):
        domain = domain.split("//", 1)[-1]
    base_url = "https://" + domain.rstrip("/") + "/daemon"

    # Check dockerfile and get directory
    dockerfile = os.path.abspath(dockerfile)
    if os.path.isfile(dockerfile):
        directory = os.path.dirname(dockerfile)
    elif os.path.isdir(dockerfile):
        directory = dockerfile
        dockerfile = os.path.join(directory, "Dockerfile")
        if not os.path.isfile(dockerfile):
            raise RuntimeError(f"No Dockerfile found in {directory!r}")
    else:
        raise RuntimeError(f"Given dockerfile not a file nor directory: {dockerfile!r}")

    # Get the client's private key, used to sign the payload
    private_key = get_private_key()

    # Get the server's time.
    # The verify=True checks the cert (default True, but let's be explicit).
    r = requests.get(base_url + "/time", verify=True)
    if r.status_code != 200:
        raise RuntimeError("Could not get server time: " + r.text)
    server_time = int(r.text)

    # Zip it up
    print("Zipping up ...")
    f = io.BytesIO()
    with zipfile.ZipFile(f, "w") as zf:
        for root, dirs, files in os.walk(directory):
            if any(x in root for x in ignore_dirs):
                continue
            for fname in files:
                filename = os.path.join(root, fname)
                fname_in_zip = os.path.relpath(filename, directory)
                if fname_in_zip != "Dockerfile":
                    zf.write(filename, fname_in_zip)
        zf.write(dockerfile, "Dockerfile")  # the deploy will simply use "Dockerfile"
    payload = f.getvalue()

    # Compose a nice little token, and a signature for it that can only be
    # produced with the private key. The public key can verify this signature
    # to confirm that we have the private key.
    fingerprint = private_key.get_id()
    token = str(server_time) + "-" + generate_uid()
    sig1 = private_key.sign(token.encode())
    sig2 = private_key.sign(payload)

    # POST to the deploy server
    url = base_url + f"/push?id={fingerprint}&token={token}"
    url += f"&sig1={quote(sig1)}&sig2={quote(sig2)}"
    print(f"Pushing ...")
    r = requests.post(url, data=payload, stream=True, verify=True)
    if r.status_code != 200:
        raise RuntimeError("Push failed: " + r.text)
    else:
        for line in r.iter_lines():
            if isinstance(line, bytes):
                line = line.decode(errors="ignore")
            print(line)



import json
import socket
import hashlib
import threading

from fastuaparser import parse_ua

from .monitor import logger


class UdpStatsReceiver(threading.Thread):
    """Thread that receives stats from UDP, send by other processes.
    Accepts (most of) statsd format, and a wee bit influxDB because that's
    what Traefik sends us.

    Processes the data and puts it into the collector.
    """

    def __init__(self, collector, port=8125):
        super().__init__()
        self._collector = collector
        self._port = port
        self.setDaemon(True)  # don't let this thread prevent shutdown
        self._stop = False

    def run(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.bind(("0.0.0.0", self._port))

        while not self._stop:
            data, addr = s.recvfrom(4096)
            try:
                self.process_data(data.decode(errors="ignore"))
            except Exception:
                pass

    def process_data(self, text):
        """Parse incoming data and put it into the collector."""
        if text.startswith("traefik"):
            group = "traefik"
            stats = self._process_data_traefik(text)
        elif text.startswith("pageview:"):
            stats = self._process_data_pageview(text)
        elif text.startswith("{"):
            stats = json.loads(text)
            group = stats.pop("group", "") or "other"
            pageview = stats.pop("pageview", None)
            if pageview:
                self._process_pageview(group, pageview)
        else:
            group = "other"
            stats = self._process_data_statsd(text)

        self._collector.put(group, stats)

    def _process_data_traefik(self, text):
        """Parsers a tiny and Traefik-specific set of influxDB."""
        stats = {}
        for line in text.splitlines():
            line = line.strip()
            if line.startswith("traefik.service.requests.total"):
                _, sep, post = line.partition(" count=")
                if sep:
                    try:
                        stats["requests|count"] = int(post.split(" ")[0])
                    except ValueError:  # pragma: no cover
                        pass
            elif line.startswith("traefik.service.connections.open"):
                _, sep, post = line.partition(" value=")
                if sep:
                    try:
                        stats["open connections|num"] = int(post.split(" ")[0])
                    except ValueError:  # pragma: no cover
                        pass
            elif line.startswith("traefik.service.request.duration"):
                _, sep, post = line.partition(" p50=")
                if sep:
                    try:
                        stats["duration|num|s"] = float(
                            post.split(" ")[0].split(",")[0]
                        )
                    except ValueError:  # pragma: no cover
                        pass
            else:
                pass  # drop it
        return stats

    def _process_pageview(self, group, headers):
        stats = {}
        try:
            stats["views|count"] = 1
            # Process referer
            referer = headers.get("referer", "")
            if referer:
                referer = referer.split("://")[-1].split("/")[0].split(":")[0]
                stats["referer|cat"] = referer
            # Use IP and user-agent to identify a user, anomimously
            # NOTE: this assumes there is a reverse proxy in front
            ip = headers.get("x-forwarded-for", "") or headers.get("x-real-ip", "")
            ua = headers.get("user-agent", "")
            lang = headers.get("accept-language", "")
            if ip and ua:
                # Get unique user string and turn it into a unique int
                client_id = ip + ua
                client_id = hashlib.md5(client_id.encode())
                client_id = abs(
                    int(client_id.hexdigest()[:14], 16)
                )  # 7 bytes fits in int64
                # Register daily visit of this user, and if its a new user, submit more
                new_user = self._collector.put_one(group, "visits|dcount", client_id)
                if new_user:
                    stats["visits|mcount"] = client_id
                    stats["client|cat"] = parse_ua(ua)  # OS and browser
                    if lang:
                        lang = lang.split(";")[0].split(",")[0].strip().lower()
                        stats["language|cat"] = lang.replace("-", " - ")
                    # todo: get country from ip
            self._collector.put(group, stats)
        except Exception as err:
            logger.error("Error processing pageview: " + str(err))

    def _process_data_statsd(self, text):
        """Process statsd data."""
        stats = {}
        for line in text.splitlines():
            parts = line.split("|")
            if len(parts) >= 2:
                name_value, type, *_ = parts
                name, value = name_value.split(":")
                if type == "c":
                    stats[name + "|count"] = int(value)
                elif type == "m":  # meter
                    stats[name + "|count"] = int(value)
                elif type == "h":  # histogram
                    stats[name + "|num"] = float(value)
                elif type == "ms":
                    stats[name + "|num|s"] = float(value) / 1000
                elif type == "g":  # we map Gauge to a num
                    stats[name + "|num"] = float(value)
                elif type == "s":
                    stats[name + "|cat"] = value
        return stats



# Yeah ... it is ugly embedding CSS like this, but this way I don't have to
# fiddle with package data when distributing. It's not that many lines!

CSS = """
body {
    font-family: Ubuntu,"Helvetica Neue",Arial,sans-serif;
    background: #333;
    color: #fff;
    padding: 0;
    margin: 0;
}
p {
    padding: 1em;
}
h1 {
    padding: 0.5em;
    text-align: center;
    color: #5af;
}
h2 {
    padding: 0.5em;
}
a, a:link, a:active, a:hover, a:visited {
    color: #bdf;
    text-decoration: none;
}
a:hover {
    color: #5af;
}
ul.links {
    font-size: 100%;
    line-height: 160%;
    padding: 0 1em;
    list-style: none;
}
ul.links > li > ul, ul.links > ul {
    list-style: none;
    padding-left: 1em;
}
.topbar {
    background: #333;
    color: #fff;
    font-size: 110%;
    margin: 0;
    margin-bottom: 6px;
}
.topbar * {
    color: #fff; /* overload link colors */
    text-decoration: none;
}
.topbar a, .topbar a:link, .topbar a:active, .topbar a:hover, .topbar a:visited {
    display: inline-block;
    padding: 0.3em;
    outline: none;
    color: #fff;
}
.topbar a:hover {
    background: rgba(128, 128, 128, 0.3);
}

.panelcontainertitle {
    text-align: center;
    background: #444;
    margin: 0;
    margin-top: 1em;
    padding: 0.3em;
}
.panelcontainer {
    min-height: 20px;
    background: #444;
    display: grid;
    grid-template-columns: auto;
    justify-items: stretch;
    align-items: stretch;
    justify-content: stretch;
    align-content: start;
    grid-auto-rows: 250px;
    padding: 4px 0px;
}

.panel {
    box-sizing: border-box;
    position: relative;
    vertical-align: top;
    background: #000;
    color: #BBB;
    box-shadow: 0 -3px 6px rgba(0, 0, 0, 0.3);
    margin: 4px;
    padding: 0;
    overflow: hidden;
}
.panel .title {
    -moz-user-select: none;
    user-select: none;
    text-align: center;
    background: #444;
    margin-bottom: 5px;
    padding: 3px;
    font-weight: bold;
}
.panel canvas {
    outline: none;
    -moz-user-select: none;
    user-select: none;
    position: absolute;
    top: 30px;
    left: 5px;
}
.panel .content {
    position: absolute;
    top: 30px;
    left: 5px;
    right: 5px;
    bottom: 5px;
    overflow: hidden;
    overflow-y: auto;
}
.panel .scrollhider {
    background: #000;
    position: absolute;
    top: 30px;
    width: 50px;
    right: 0px;
    bottom: 0px;

}
.panel .content table {
    width: 100%;
    text-align: left;
    border-collapse: collapse;
    font-size: 85%;
}
table.info {
    text-align: left;
    border-collapse: collapse;
}
.panel .content table td, table.info td {
    padding-left: 0.7em;
}
.panel .content table tr, table.info tr {
    background-color: rgba(100, 100, 100, 0.2);
}
.panel .content table tr:nth-child(even), table.info tr:nth-child(even) {
    background-color: rgba(160, 160, 160, 0.2);
}

@media screen and (min-height: 600px) {
    .panelcontainer {
       grid-auto-rows: 300px;
    }
}
@media screen and (min-width: 600px) {
    .panelcontainer {
       grid-template-columns: auto auto;
       padding: 4px 4px;
    }
}
@media screen and (min-width: 1200px) {
    .panelcontainer {
       grid-template-columns: auto auto auto;
    }
}
@media screen and (min-width: 1600px) {
    .panelcontainer {
       grid-template-columns: auto auto auto auto;
    }
}
"""



import os

from .. import __traefik_version__
from ..utils import dockercall
from ._auth import load_config


# Notes:
# * Traefik checks whether the domain actually resolves to itself,
#   see disablepropagationcheck


def restart_router():
    """Restart the Traefik docker container. You can run this after
    updating the config (~/_mypaas/traefik.toml or staticroutes.toml)
    or to update Traefik after updating MyPaas. Your PAAS will be
    offline for a few seconds.
    """

    image_name = "traefik:" + __traefik_version__

    print(f"Pulling Docker image: {image_name}")
    dockercall("pull", image_name)

    print("Stopping and removing current Traefik container (ignore errors)")
    dockercall("stop", "traefik", fail_ok=True)
    dockercall("rm", "traefik", fail_ok=True)

    print("Launching new Traefik container")
    cmd = ["run", "-d", "--restart=always"]
    traefik_dir = os.path.expanduser("~/_mypaas")
    cmd.extend(["--network=host", "-p=80:80", "-p=443:443"])
    cmd.append("--volume=/var/run/docker.sock:/var/run/docker.sock")
    cmd.append(f"--volume={traefik_dir}/traefik.toml:/traefik.toml")
    cmd.append(f"--volume={traefik_dir}/acme.json:/acme.json")
    cmd.append(f"--volume={traefik_dir}/staticroutes.toml:/staticroutes.toml")
    cmd.append(f"--env=MYPAAS_SERVICE=traefik")
    cmd.append(f"--env=MYPAAS_CONTAINER=traefik")
    cmd.extend(["--name=traefik", image_name])
    dockercall(*cmd)


def init_router():
    """
    Prepare the system for running Traefik (Docker network and config).
    Running this again will reset Traefik "to factory defaults".
    """

    # Get config
    traefik_dir = os.path.expanduser("~/_mypaas")
    config = load_config()

    # Make sure there is an acme.json with the right permissions
    acme_filename = os.path.join(traefik_dir, "acme.json")
    if os.path.isfile(acme_filename):
        print(f"Leaving {acme_filename} (containing certificates) as it is.")
    else:
        print(f"Creating {acme_filename} (for certificates)")
        with open(os.path.join(traefik_dir, "acme.json"), "wb"):
            pass
    os.chmod(os.path.join(traefik_dir, "acme.json"), 0o600)

    # Create the static config
    print("Writing Traefik config")
    text = traefik_config.replace("EMAIL", config["init"]["email"])
    with open(os.path.join(traefik_dir, "traefik.toml"), "wb") as f:
        f.write(text.encode())

    # Create the file-provider's config
    print("Writing Traefik static routes")
    text = traefik_staticroutes.replace("PAAS_DOMAIN", config["init"]["domain"])
    text = text.replace("WEB_CREDENTIALS", config["init"]["web_credentials"])
    with open(os.path.join(traefik_dir, "staticroutes.toml"), "wb") as f:
        f.write(text.encode())


traefik_config = """
# Traefik config file.
# https://docs.traefik.io/reference/static-configuration/file/
#
# Traefik must be restarted for changes to take effect.
# This file will be overwritten when you run `mypaas server init`.


[log]
  level = "WARN"

# Define entrypoints: what ports Traefik listens on
[entryPoints]
  [entryPoints.web]
    address = ":80"
  [entryPoints.web-secure]
    address = ":443"

# Define providers for dynamic config. We use Docker and a file
[providers.docker]
  endpoint = "unix:///var/run/docker.sock"
  network = "mypaas-net"
  watch = true
  exposedByDefault = false
  useBindPortIP = false
[providers.file]
  filename = "/staticroutes.toml"
  watch = true

# Enable dashboard
[api]
  dashboard = true

# Enable Let's Encrypt. Use EC256 because https://github.com/almarklein/mypaas/pull/24
[certificatesResolvers.default.acme]
  email = "EMAIL"
  keyType = "EC256"
  storage = "acme.json"
  [certificatesResolvers.default.acme.httpchallenge]
    entrypoint = "web"

# Process metrics (use the influxDB protocol, because it sends aggregates)
[metrics]
  [metrics.influxDB]
    address = "127.0.0.1:8125"
    addEntryPointsLabels = false
    addServicesLabels = true
    pushInterval = "1s"

""".lstrip()


traefik_staticroutes = """
# Trafic config for statically defined routes and middleware.
# Traefik should update automatically when changed are made (without restart).
# This file will be overwritten when you run `mypaas server init`.

# The Traefik dashboard
[http.routers.api]
  rule = "Host(`PAAS_DOMAIN`) && (PathPrefix(`/dashboard`) || PathPrefix(`/api`))"
  entrypoints = ["web-secure"]
  service = "api@internal"
  middlewares = ["auth"]
  [http.routers.api.tls]
    certresolver = "default"
    options= "intermediate"

# The routing for mypaas daemon
[http.routers.mypaas-daemon-router]
  rule = "Host(`PAAS_DOMAIN`) && PathPrefix(`/daemon`)"
  entrypoints = ["web-secure"]
  service = "mypaas-daemon"
  [http.routers.mypaas-daemon-router.tls]
    certresolver = "default"
    options= "intermediate"
[http.services.mypaas-daemon.loadBalancer]
    [[http.services.mypaas-daemon.loadBalancer.servers]]
      url = "http://127.0.0.1:88"

[http.middlewares]
  [http.middlewares.https-redirect.redirectscheme]
    scheme = "https"
  [http.middlewares.hsts-header.headers]
    [http.middlewares.hsts-header.headers.customResponseHeaders]
      Strict-Transport-Security = "max-age=63072000"

# Better security. For details see https://github.com/almarklein/mypaas/pull/25
[tls.options]
  [tls.options.intermediate]
    minVersion = "VersionTLS12"
    cipherSuites = [
      "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256",
      "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256",
      "TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384",
      "TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384",
      "TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305",
      "TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305"
    ]

# You can update/add users here and then 'mypaas server restart traefik'
# Create a password hash using e.g. 'openssl passwd -apr1'
[http.middlewares.auth.basicAuth]
  users = ["WEB_CREDENTIALS"]

""".lstrip()



"""
The CLI for mypaas.
"""

import sys

import mypaas
from mypaas import __version__, __traefik_version__


def server_help():
    """Show this help message and exit."""
    print(server_docs)


def client_help():
    """Show this help message and exit."""
    print(client_docs)


server_help.__name__ = client_help.__name__ = "help"


def version():
    """Print version."""
    print(f"mypaas v{__version__} (using Traefik v{__traefik_version__})")


def _make_func_dict_and_docs(*args):
    funcs = {}
    description = "usage: mypaas command [arguments]"
    description += "\n\n" + mypaas.__doc__.strip() + "\n\n"

    for func in args:
        if isinstance(func, str):
            description += func + "\n\n"  # header
        else:
            funcs[func.__name__] = func
            funcs[func.__name__.replace("_", "-")] = func
            co = func.__code__
            arg_names = " ".join(x.upper() for x in co.co_varnames[: co.co_argcount])
            description += "    " + func.__name__ + " " + arg_names + "\n"
            doc = "    " + func.__doc__.strip()
            description += doc.replace("    ", "        ") + "\n"

    return funcs, description


server_funcs, server_docs = _make_func_dict_and_docs(
    "Commands to run at the PaaS server:",
    version,
    server_help,
    mypaas.server.init,
    mypaas.server.restart,
    mypaas.server.deploy,
    mypaas.server.schedule_reboot,
)


client_funcs, client_docs = _make_func_dict_and_docs(
    "This is the client CLI. Use 'mypaas server ..' for server commands.\n"
    + "Commands to run at your work machine:",
    version,
    client_help,
    mypaas.client.key_init,
    mypaas.client.key_gen,
    mypaas.client.key_get,
    mypaas.client.push,
)


def main(argv=None):

    assert sys.version_info.major == 3, "This script needs to run with Python 3."

    # Get CLI args and determine whether we run as server
    if argv is None:
        argv = sys.argv[1:]
    is_server = False
    if argv and argv[0] == "server":
        is_server = True
        argv = argv[1:]
    elif sys.argv[0].endswith(("-server", "_server")):
        is_server = True
    if not argv:
        argv = ["help"]

    # Get function to call
    if is_server:
        if argv[0] in server_funcs:
            func = server_funcs[argv[0]]
        elif argv[0] in client_funcs:
            sys.exit(f"MyPaas subcommand {argv[0]} is a client (not server) command.")
        else:
            sys.exit(f"Invalid use of mypaas: {argv}")
    else:
        if argv[0] in client_funcs:
            func = client_funcs[argv[0]]
        elif argv[0] in server_funcs:
            sys.exit(f"MyPaas subcommand {argv[0]} is a server (not client) command.")
        else:
            sys.exit(f"Invalid use of mypaas: {argv}")

    # Call it
    try:
        func(*argv[1:])
    except RuntimeError as err:
        # Inside the functions, RunTimeError is raised in situations
        # that can sufficiently be described with the error message.
        # Other exceptions fall through, and their traceback is
        # printed.
        sys.exit(str(err))


if __name__ == "__main__":
    main()



import random
import subprocess


def input_ask_bool(s):
    while True:
        x = input(s)
        if x.lower() in ("y", "yes"):
            return True
        elif x.lower() in ("n", "no"):
            return False
        else:
            print("Invalid answer. Please answer with one of: 'y', 'yes', 'n', 'no'.\n")


def input_ask_int(s, valid_numbers):
    while True:
        x = input(s)
        try:
            n = int(x)
        except ValueError:
            pass
        else:
            if n in valid_numbers:
                return n
        print("Invalid answer. Please provide a valid number.\n")


def generate_uid(n=8):
    """Generate a unique id in the form of an 8-char string"""
    chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    nchars = len(chars)  # 52, so 52**8 => 53459728531456 possibilities
    return "".join([chars[int(random.random() * nchars)] for i in range(n)])


def dockercall(*args, fail_ok=False):
    cmd = ["docker"]
    cmd.extend(args)
    try:
        output_bytes = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        return output_bytes.decode(errors="ignore").strip()
    except subprocess.CalledProcessError as err:
        output = err.output.decode(errors="ignore").strip()
        if fail_ok:
            return output
        else:
            header = "=======> Calling [docker, " + str(cmd)[1:] + "\n"
            raise RuntimeError(header + output + "\n=======> Error in Docker call!")



"""
Main script to run the stats server.
"""

import os

import asgineer

try:
    from stats import StatsCollector, UdpStatsReceiver
    from stats import stats_handler
except ImportError:
    from mypaas.stats import StatsCollector, UdpStatsReceiver
    from mypaas.stats import stats_handler


# Create a stats collector
db_dir = os.path.expanduser("~/_stats")
collector = StatsCollector(db_dir)

# Start a thread that receives stats via udp and puts it into the collector
udp_stats_receiver = UdpStatsReceiver(collector)
udp_stats_receiver.start()


@asgineer.to_asgi
async def main_handler(request):
    return await stats_handler(request, collector)


if __name__ == "__main__":
    asgineer.run(main_handler, "uvicorn", "0.0.0.0:80", log_level="warning")



import os
import time

from ..utils import PublicKey

import toml


server_key_filename = "~/_mypaas/authorized_keys"
config_filename = "~/_mypaas/config.toml"

last_key_read = 0
_authorized_keys = {}


def get_public_key(fingerprint):
    """Get the public key for the given fingerprint"""
    # Read the keys from the filesystem at most once each few seconds,
    # to prevent attacks on the auth service.
    global last_key_read
    if last_key_read < time.time() - 5:
        last_key_read = time.time()
        _authorized_keys.clear()
        _authorized_keys.update(get_authorized_keys(server_key_filename))

    return _authorized_keys.get(fingerprint, None)


def get_authorized_keys(filename):
    """Read the authorized public keys from the file system.
    Returns a dict of PublicKey objects.
    """
    if filename.startswith("~"):
        filename = os.path.expanduser(filename)
    if not os.path.isfile(filename):
        return {}

    with open(filename, "rb") as f:
        text = f.read().decode(errors="ignore")

    keys = {}
    for line in text.split("\n"):
        line = line.strip()
        if not line or line.startswith("#"):
            continue
        try:
            key = PublicKey.from_str(line)
        except Exception:
            print(f"Does not look like a public key: {line}")
        else:
            keys[key.get_id()] = key
    return keys


def load_config():
    """Load config from disk."""
    filename = os.path.expanduser(config_filename)
    try:
        with open(filename, "rb") as f:
            return toml.loads(f.read().decode())
    except Exception:
        return {"init": {}, "env": {}}


def save_config(config):
    """Save config to disk."""
    filename = os.path.expanduser(config_filename)
    with open(filename, "wb") as f:
        f.write(toml.dumps(config).encode())



"""
The mypaas.utils subpackage contains utilities for both server and client.

Dependencies: cryptography
"""

# flake8: noqa

from ._utils import input_ask_bool, input_ask_int, generate_uid, dockercall
from ._crypto import PublicKey, PrivateKey



import os
import shutil

from ._deploy import deploy  # noqa

import toml


THIS_DIR = os.path.abspath(os.path.dirname(__file__))
STATS_LIB_DIR = os.path.join(os.path.dirname(THIS_DIR), "stats")


def restart_stats():
    """Restart the stats server."""

    # Get config
    config_filename = os.path.expanduser("~/_mypaas/config.toml")
    with open(config_filename, "rb") as f:
        config = toml.loads(f.read().decode())

    deploy_dir = os.path.expanduser("~/_mypaas/stats_deploy_cache")
    dockerfile = DOCKERFILE.replace("PAAS_DOMAIN", config["init"]["domain"])

    try:
        # Put required files in deploy dir
        shutil.rmtree(deploy_dir, ignore_errors=True)
        shutil.copytree(STATS_LIB_DIR, os.path.join(deploy_dir, "stats"))
        with open(os.path.join(deploy_dir, "Dockerfile"), "wb") as f:
            f.write(dockerfile.encode())
        # Deploy
        deploy(deploy_dir)
    finally:
        shutil.rmtree(deploy_dir, ignore_errors=True)


DOCKERFILE = """
# mypaas.service=stats
# mypaas.url=https://PAAS_DOMAIN/
# mypaas.port=80
# mypaas.publish=127.0.0.1:8125:8125/udp
# mypaas.volume=/root/_stats:/root/_stats
# mypaas.scale=0

FROM ubuntu:20.04

RUN apt update \
    && apt install -y python3-psutil python3-pip \
    && pip3 --no-cache-dir install pip --upgrade \
    && pip3 --no-cache-dir install uvicorn uvloop httptools \
    && pip3 --no-cache-dir install asgineer>=0.8 itemdb pscript fastuaparser

WORKDIR /root
COPY . .
CMD ["python3", "-m", "stats"]
""".lstrip()



"""
Code related to key pairs at the client.
"""

import os
import getpass

import pyperclip

from ..utils import input_ask_bool, input_ask_int, PrivateKey

standard_key_filename = os.path.normpath(os.path.expanduser("~/.ssh/id_rsa"))
default_key_filename = os.path.normpath(os.path.expanduser("~/.ssh/mypaas_rsa"))

os.makedirs(os.path.expanduser("~/.ssh"), exist_ok=True)

private_text = """
It's important to keep the private key secret: store it in a safe place
and nowhere else. You can always create a new key if needed.
""".strip()

public_text = """
The public key is used by the server to confirm that you have the
private key. You can safely share the public key; it is not a secret.
The public key is now on the clipboard. SSH into your paas server
and add it to ~/_mypaas/authorized_keys
""".rstrip()

OPTIONS = [
    f"0. Keep as it is",
    f"1. Generate a new RSA keypair (and store it in '{default_key_filename}')",
    f"2. Use an existing RSA key file",
    f"3. Use the standard RSA key '{standard_key_filename}'",
]


def get_key_filename_and_text():
    """Get the filename that contains the RSA key, and the containing text.
    We use one specific file location, but it may point to another RSA file.
    """
    filename = default_key_filename

    if not os.path.isfile(filename):
        raise RuntimeError("RSA keyfile does not exist. Create a new keypair.")

    with open(filename, "rb") as f:
        text = f.read().decode().strip()

    if text.startswith("file://"):
        filename = text.split("//", 1)[-1]
        if not os.path.isfile(filename):
            raise RuntimeError(
                f"Referenced keyfile '{filename}' does not exist. Create new keypair."
            )
        with open(filename, "rb") as f:
            text = f.read().decode().strip()

    return filename, text


def key_init():
    """Setup a keypair to authorize this machine to a MyPaas server.
    The private key is stored on this machine and should be kept secret.
    You can also choose to use an existing RSA key.
    """

    # Get current status
    try:
        filename, text = get_key_filename_and_text()
        status = f"Using RSA keyfile: '{filename}'"
        current_status_ok = True
    except RuntimeError as err:
        status = str(err)
        current_status_ok = False
    print("Current client status:\n    " + status)
    print()

    # Get options, and remove the ones that do not make sense.
    options = OPTIONS.copy()
    if not current_status_ok:
        options.pop(0)  # Doing nothing does not make sense when the state sucks
    if not os.path.isfile(standard_key_filename):
        options.pop(-1)  # Standard SSH file not there

    # Ask user what to do
    valid_choices = [int(option[0]) for option in options]
    choice = input_ask_int("\n".join(options) + "\nYour choice> ", valid_choices)

    if choice == 0:
        print("No action required.")
        return

    elif choice == 1:
        filename = default_key_filename
        # If there already is a key, we maybe should not overwrite it
        if os.path.isfile(filename) and current_status_ok:
            if not input_ask_bool(
                f"A key is already present at '{filename}'. Overwrite? (y/n)> "
            ):
                return
        # Generate the keypair
        print("\nGenerating keypair ... ", end="")
        private_key = PrivateKey.generate()
        print("Done.")
        # Store the private key, with a passphrase
        print("It is strongly recommended to protect your key with a passprase.")
        print("You'll be typing it in a lot; keep it simple (but not too short).")
        pp = getpass.getpass("Your passphrase to protect the key: ")
        with open(filename, "wb") as f:
            f.write(private_key.to_str(pp).encode())
        print(f"Your private key is now stored at '{filename}'.")

    elif choice == 2 or choice == 3:
        # Select and check filename
        if choice == 3:
            filename = standard_key_filename
        else:
            filename = input("Path to RSA file> ")
        if not os.path.isfile(filename):
            raise RuntimeError(f"File does not exist: '{filename}'")
        # Write the proxy
        with open(default_key_filename, "wb") as f:
            f.write(("file://" + filename).encode())
        print(f"Now using private key from '{filename}'.")

    else:
        raise RuntimeError("Woops, Don't know what to do.")  # should be unreachable


def key_gen():
    """
    Generate (but not store) a generic (passwordless) SSH keypair,
    e.g. to use in CI/CD.
    """

    # Generate the keypair
    print("\nGenerating keypair ... ", end="")
    private_key = PrivateKey.generate()
    public_key = private_key.get_public_key()
    print("Done.")

    # Copy private key
    print(private_text)
    print("")
    input("Hit enter to copy the private key to the clipboard. > ")
    pyperclip.copy(private_key.to_str(None).replace("\n", "_"))
    try:
        print("Now paste the key to its destnation, ")
        input("and then hit enter again to clear the clipboard. > ")
    finally:
        pyperclip.copy("")

    # Copy public key
    _show_public_key(public_key)


def key_get():
    """
    Get the public key corresponding to the private key on this machine.
    This public key can be shared. The key is printed and copied to the
    clipboard.
    """
    private_key = get_private_key()
    public_key = private_key.get_public_key()
    _show_public_key(public_key)


def get_private_key():
    """Get the private key, so our commands like `push` can authenticate."""
    keyname = "MYPAAS_PRIVATE_KEY"

    if os.getenv(keyname, ""):
        text = os.environ[keyname]
        filename = keyname
        pp = ""
    else:
        filename, text = get_key_filename_and_text()
        pp = getpass.getpass(f"Passphrase for key '{filename}': ")

    try:
        return PrivateKey.from_str(text, pp)
    except Exception as err:
        raise RuntimeError(f"Could not load key from {filename}: {str(err)}")


def _show_public_key(public_key):
    pyperclip.copy(public_key.to_str())
    print()
    print(public_key.to_str())
    print(public_text)
    print("\nKeypair fingerprint: " + public_key.get_id())



"""
This mypaas.server subpackage provides a CLI to run at the PaaS server
to manage it.

Dependencies: asgineer
"""

# flake8: noqa

from ._init import init, restart
from ._traefik import init_router, restart_router
from ._deploy import deploy, get_deploy_generator
from ._auth import get_public_key
from ._daemon import restart_daemon
from ._stats import restart_stats
from ._reboot import schedule_reboot



import json
import time
import socket
import logging
import threading

import psutil


logger = logging.getLogger("mypaas.daemon")

stats_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)


class SystemStatsProducer(threading.Thread):
    """Thead that produces measurements on the system and mypaas services,
    and sends these stats to the mypaas stats server.
    Currently measuring CPU, RAM and disk.
    """

    def __init__(self):
        super().__init__()
        self.setDaemon(True)
        self._stop = False
        self._service_processes = {}
        self._create_times = {}

    def run(self):
        t = time.time()
        time1 = t + 1
        time10 = t + 3  # the first 10-tick comes sooner

        while not self._stop:
            time.sleep(0.05)
            t = time.time()

            if t > time1:
                time1 = t + 1
                try:
                    self._do_each_1_seconds()
                except Exception:  # pragma: no cover
                    pass

            if t > time10:
                time10 = t + 10
                try:
                    self._do_each_10_seconds()
                except Exception:  # pragma: no cover
                    pass

    def _send(self, stat):
        stats_socket.sendto(json.dumps(stat).encode(), ("localhost", 8125))

    def _do_each_1_seconds(self):
        self._measure_stats_of_system()
        self._measure_stats_of_services()

    def _do_each_10_seconds(self):
        self._measure_system_disk_usage()
        self._collect_services()
        self._detect_startups()

    def _measure_stats_of_system(self):
        try:
            stat = {
                "group": "system",
                "cpu|num|%": max(0.01, psutil.cpu_percent()),
                "mem|num|iB": psutil.virtual_memory().used,
            }
            self._send(stat)
        except Exception as err:  # pragma: no cover
            logger.error("Failed to send system measurements: " + str(err))

    def _measure_system_disk_usage(self):
        try:
            disk = psutil.disk_usage("/").used
            stat = {"group": "system", "disk|num|iB": disk}
            self._send(stat)
        except Exception as err:  # pragma: no cover
            logger.error("Failed to send system measurements: " + str(err))

    def _measure_stats_of_services(self):
        for container_name, p in self._service_processes.items():
            try:
                stat = {
                    "group": container_name,
                    "cpu|num|%": max(0.01, p.cpu_percent()),
                    "mem|num|iB": p.memory_info().rss,
                }
                self._send(stat)
            except Exception as err:  # pragma: no cover
                logger.error(
                    f"Failed to send {container_name} measurements: " + str(err)
                )

    def _detect_startups(self):
        for container_name, p in self._service_processes.items():
            # Get process creation time and our stored creation time
            try:
                create_time = p.create_time()
            except Exception:
                continue  # p.create_time may not be available on each OS
            ref_create_time = self._create_times.get(container_name, 0)
            # If it's different and uptime is less than a minute, assume its a startup
            # Note that we can get false positives if *this* process restarts within
            # a minute after another process does.
            if ref_create_time != create_time:
                self._create_times[container_name] = create_time
                uptime = time.time() - create_time
                if uptime < 60:
                    try:
                        stat = {
                            "group": container_name,
                            "startup|count": 1,
                        }
                        self._send(stat)
                    except Exception as err:  # pragma: no cover
                        logger.error(
                            f"Failed to send {container_name} startup: " + str(err)
                        )

    def _collect_services(self):
        try:
            processes = {}
            for p in psutil.process_iter():
                try:
                    container_name = p.environ().get("MYPAAS_CONTAINER", "")
                    if container_name:
                        processes[container_name] = p
                except Exception:  # pragma: no cover
                    pass
            self._service_processes = processes
        except Exception as err:  # pragma: no cover
            logger.error("Failed to collect service processes: " + str(err))



"""
Entrypoint for the MyPaas daemon.
"""

import asgineer

from mypaas.daemon import main_handler, SystemStatsProducer


@asgineer.to_asgi
async def main(request):
    return await main_handler(request)


stats_producer = SystemStatsProducer()
stats_producer.start()


if __name__ == "__main__":
    # Host on localhost. Traefik will accept https connections and route via
    # localhost:88. Don't allow direct acces, or we have insecure connection.
    asgineer.run(main, "uvicorn", "localhost:88", log_level="warning")



"""
MyPaas is a tool that makes it easy to run a platform as a service (PaaS)
on your own VM or hardware. It combines Traefik and Docker, enabling free
automatic https (via Let's Encrypt) and deployments via dockerfiles.
"""

# flake8: noqa


__version__ = "0.5.5"
__traefik_version__ = "2.4.7"

version_info = tuple(map(int, __version__.split(".")))

from . import utils
from . import server
from . import client



"""
The mypaas.client subpackage contains the logic for the commands to run at the client.

Dependencies: pyperclip, requests
"""

# flake8: noqa

from ._keys import key_init, key_gen, key_get
from ._push import push



"""
Python code that will be transpiled to JS to implement the client side.
"""

from pscript.stubs import window, document, undefined, Math, Date  # JS
from pscript.stubs import data_per_db, text_color  # are made available


panels = []


# %% Button callbacks


def toggle_utc():
    info = get_hash_info()
    if info.get("utc", False):
        info.pop("utc")
    else:
        info["utc"] = True
    return refresh(None, info)


def toggle_columns():
    info = get_hash_info()
    columns = info.get("columns", 0)
    if not columns:
        if window.document.body.clientWidth >= 1200:
            info["columns"] = 2
        else:
            info["columns"] = 1
    else:
        info.pop("columns")
    return refresh(None, info)


def update_range(action=""):

    ndays = window.ndays
    daysago = window.daysago
    if action == "zoomout":
        if ndays < 4:
            ndays += 1
        elif ndays < 10:
            ndays += 2
        elif ndays < 30:
            ndays += 5
        else:
            ndays += 10
    elif action == "zoomin":
        if ndays <= 4:
            ndays -= 1
        elif ndays <= 10:
            ndays -= 2
        elif ndays <= 30:
            ndays -= 5
        else:
            ndays -= 10
        ndays = max(1, ndays)
    elif action == "older":
        daysago += ndays
    elif action == "newer":
        daysago -= ndays

    info = get_query_info()
    info["ndays"] = ndays
    if daysago > 0:
        info["daysago"] = daysago
    else:
        info.pop("daysago", None)
    return refresh(info, None)


def refresh(self, query_info=None, hash_info=None):
    if query_info is None:
        query_info = get_query_info()
    if hash_info is None:
        hash_info = get_hash_info()

    url = window.location.origin + window.location.pathname
    encode_uri_component = window.encodeURIComponent
    if query_info:
        url += "?" + "&".join(
            [key + "=" + encode_uri_component(val) for key, val in query_info.items()]
        )
    if True:
        url += "#" + "&".join(
            [key + "=" + encode_uri_component(val) for key, val in hash_info.items()]
        )

    if url == window.location.href:
        window.location.reload()
    else:
        window.location.href = url
    return undefined


def panel_sort_func(x):
    t = x.split("|")[1]
    if t:
        t = {"num": "anum", "cat": "zcat"}.get(t, t)
    return (t + "|" + x).lower()


# %%


def on_init():

    for dbname, data in data_per_db.items():

        # Create panel container (and a title)
        title_el = document.createElement("div")
        container_el = document.createElement("div")
        title_el.innerText = dbname  # .replace("_", " ")
        title_el.classList.add("panelcontainertitle")
        container_el.classList.add("panelcontainer")
        document.body.appendChild(title_el)
        document.body.appendChild(container_el)

        if dbname == "system" and window.info:
            panels.append(InfoPanel(container_el, dbname, "info", "system info"))

        # Collect panel types
        panel_kinds = {}
        for i in range(len(data)):
            aggr = data[i]
            for key in aggr.keys():
                panel_kinds[key] = True

        # Sort the panel types - count, dcount, num, cat
        panel_kinds = panel_kinds.keys()
        panel_kinds.sort(key=panel_sort_func)

        # Create panels
        for i in range(len(panel_kinds)):
            key = panel_kinds[i]
            # Select panel class
            key_parts = key.split("|")
            if len(key_parts) == 2:
                name, type = key_parts
                unit = ""
            elif len(key_parts) == 3:
                name, type, unit = key_parts
            else:
                continue
            if type == "time":
                continue  # skip time info
            elif type == "count":
                title = "# " + name
                Cls = CountPanel  # noqa: N806
            elif type == "dcount":
                title = "# daily " + name
                Cls = DailyCountPanel  # noqa: N806
            elif type == "mcount":
                title = "# monthly " + name
                Cls = MonthlyCountPanel  # noqa: N806
            elif type == "cat":
                title = name + "'s"
                Cls = CategoricalPanel  # noqa: N806
            elif type == "num":
                title = name
                Cls = NumericalPanel  # noqa: N806
            else:
                window.console.warn(f"Don't know what to do with {key}")
                continue
            if unit:
                title = title + " " + unit
            # Create panel
            panel = Cls(container_el, dbname, key, title, unit)
            panels.append(panel)

    on_hash_change()  # calls on_resize()


def on_resize():
    window.setTimeout(_on_resize, 1)


def get_query_info():
    url = window.location.href
    q = ""
    if "?" in url:
        q = window.location.href.split("?", 1)[-1].split("#")[0]
    return get_dict_from_hash_or_query(q)


def get_hash_info():
    return get_dict_from_hash_or_query(window.location.hash.lstrip("#"))


def get_dict_from_hash_or_query(s):
    info = {}
    for s in s.split("&"):
        key, _, val = s.partition("=")
        if key and val:
            val = window.decodeURIComponent(val)
            if val.lower() == "true":
                val = True
            elif val.lower() == "false":
                val = False
            elif val in "0123456789":
                val = int(val)
            info[key] = val
        elif s:
            info[s] = True
    return info


def on_hash_change():
    info = get_hash_info()
    containers = document.getElementsByClassName("panelcontainer")

    columns = int(info.get("columns", "")) or 0
    if columns > 0:
        grid_template_columns = "auto ".repeat(columns)
    else:
        grid_template_columns = None

    height = int(info.get("height", "")) or 0
    if height > 0:
        grid_auto_rows = height + "px"
    else:
        grid_auto_rows = None

    for i in range(len(containers)):
        containers[i].style.gridAutoRows = grid_auto_rows
        containers[i].style.gridTemplateColumns = grid_template_columns

    on_resize()


def _on_resize():
    for panel in panels:
        if panel.canvas:
            # Get dimensions
            w = panel.node.clientWidth - 10
            h = panel.node.clientHeight - 35
            pixel_ratio = get_pixel_ratio(panel.canvas.getContext("2d"))
            # Set dimensions
            panel.canvas.style.width = w + "px"
            panel.canvas.style.height = h + "px"
            panel.canvas.width = w * pixel_ratio
            panel.canvas.height = h * pixel_ratio
            # Set some info on the object
            panel.pixel_ratio = pixel_ratio
            panel.width = w
            panel.height = h
        if panel.draw:
            panel.draw()


def get_pixel_ratio(ctx):
    """Get the ratio of logical pixel to screen pixel."""
    PSCRIPT_OVERLOAD = False  # noqa

    dpr = window.devicePixelRatio or 1
    bsr = (
        ctx.webkitBackingStorePixelRatio
        or ctx.mozBackingStorePixelRatio
        or ctx.msBackingStorePixelRatio
        or ctx.oBackingStorePixelRatio
        or ctx.backingStorePixelRatio
        or 1
    )
    return dpr / bsr


def _create_tick_units():
    # Create tick units
    tick_units = []
    for e in range(-14, 14):
        for i in [10, 20, 25, 50]:
            tick_units.append(i * 10 ** e)
    return tick_units


_tick_units = _create_tick_units()


# def split_group(s, sep):
#     group, _, sub = s.partition(sep)
#     if len(sub) == 0:
#         return "", group
#     else:
#         return group, sub


class BasePanel:
    def __init__(self, container, dbname, key, title, unit):
        self.dbname = dbname
        self.key = key
        self.title = title
        self.unit = unit

        self.node = document.createElement("div")
        self.node.classList.add("panel")
        container.appendChild(self.node)

        self.titlenode = document.createElement("div")
        self.titlenode.classList.add("title")
        self.titlenode.innerText = title
        self.node.appendChild(self.titlenode)


class InfoPanel(BasePanel):
    def __init__(self, *args):
        super().__init__(*args)
        self.content = document.createElement("div")
        self.content.classList.add("content")
        self.node.appendChild(self.content)

        hider = document.createElement("div")
        hider.classList.add("scrollhider")
        self.node.appendChild(hider)

        self._create()

    def _create(self):
        PSCRIPT_OVERLOAD = False  # noqa
        if not window.info:
            return
        lines = []
        lines.append("<table>")
        for key, value in window.info.items():
            lines.append(f"<tr> <td>{key}</td> <td>{value}</td> </tr>")
        lines.append("</table>")
        self.content.innerHTML = "\n".join(lines)


class CategoricalPanel(InfoPanel):
    def _create(self):
        PSCRIPT_OVERLOAD = False  # noqa

        key = self.key

        # First aggregate
        data = data_per_db[self.dbname]
        totalcount = 0
        rows = {}
        for i in range(len(data)):
            aggr = data[i]
            meas = aggr.get(key, {})
            for k, v in meas.items():
                rows[k] = rows.get(k, 0) + v
                totalcount += v

        # Group so we can sort in a grouped fashion
        groups = {}
        group_counts = {}
        for key, count in rows.items():
            group, _, subkey = key.partition(" - ")
            groups.setdefault(group, []).append((subkey, count))
            group_counts[group] = group_counts.get(group, 0) + count
        group_counts = [(k, v) for k, v in group_counts.items()]

        # Sort groups and items inside the groupd
        group_counts.sort(key=lambda x: -x[1])
        for subs in groups.values():
            subs.sort(key=lambda x: -x[1])

        lines = []
        lines.append("<table>")
        for group, _ in group_counts:
            for sub, count in groups[group]:
                key = group + " - " + sub
                key = key.strip(" -")
                pct = 100 * count / totalcount
                lines.append(
                    f"<tr> <td>{pct:0.0f}%</td> <td>{count}</td> <td>{key}</td> </tr>"
                )
        lines.append("</table>")

        self.content.innerHTML = "\n".join(lines)


class PlotPanel(BasePanel):

    _values_are_integer = False

    def __init__(self, *args):
        super().__init__(*args)
        self.canvas = document.createElement("canvas")
        self.node.appendChild(self.canvas)

    def _draw_text(self, ctx, text, x, y, angle=0):
        PSCRIPT_OVERLOAD = False  # noqa
        ctx.save()
        ctx.translate(x, y)
        ctx.scale(1, -1)
        ctx.rotate(angle)
        ctx.fillText(text, 0, 0)
        ctx.restore()

    def _get_min_max(self):
        return 0, 1

    def _get_ticks(self, scale, mi, ma, min_tick_dist=40):
        PSCRIPT_OVERLOAD = False  # noqa
        # Inspired from flexx' PlotWidget, which took inspirartion from visvis

        # Get tick multipliers and unit modifier
        if self.unit == "iB":
            if ma >= 2 ** 30:
                mult, unit = 1 / 2 ** 30, "G"
            elif ma >= 2 ** 20:
                mult, unit = 1 / 2 ** 20, "M"
            elif ma >= 2 ** 10:
                mult, unit = 1 / 2 ** 10, "K"
            else:
                mult, unit = 1, ""
        else:
            if ma >= 10_000_000_000:
                mult, unit = 1 / 1_000_000_000, "G"
            elif ma >= 10_000_000:
                mult, unit = 1 / 1_000_000, "M"
            elif ma >= 10000:
                mult, unit = 1 / 1000, "K"
            elif ma < 0.0001:
                mult, unit = 1_000_000, "u"
            elif ma < 0.1:
                mult, unit = 1000, "m"
            else:
                mult, unit = 1, ""
        if self.unit in ("iB", "s"):
            title = self.title.replace(" " + self.unit, " " + unit + self.unit)
            self.titlenode.innerText = title
            unit = ""
        # Get tick unit
        is_int = self._values_are_integer
        for tick_unit in _tick_units:
            if is_int and str(tick_unit).indexOf(".") >= 0:
                continue
            if tick_unit * scale / mult >= min_tick_dist:
                break
        else:
            return []
        # Calculate tick values
        first_tick = Math.ceil(mi * mult / tick_unit) * tick_unit
        last_tick = Math.floor(ma * mult / tick_unit) * tick_unit
        ticks = {}
        t = first_tick  # t does not mean time here!
        while t <= last_tick:
            ticks[t / mult] = t
            t += tick_unit
        # Stringify
        for realt, t in ticks.items():
            if t == 0:
                s = "0"
            elif mult == 1 and is_int:
                s = str(int(t))
            else:
                s = t.toPrecision(4)  # t is already multiplied
                if "." in s:
                    while len(s) > 5 and s.endsWith("0"):
                        s = s[:-1]
            ticks[realt] = s + unit
        return ticks

    def draw(self):
        PSCRIPT_OVERLOAD = False  # noqa

        ctx = self.canvas.getContext("2d")

        # Prepare hidpi mode for canvas  (flush state just in case)
        for i in range(4):
            ctx.restore()
        ctx.save()
        ctx.scale(self.pixel_ratio, self.pixel_ratio)

        # Flip y-axis
        ctx.scale(1, -1)
        ctx.translate(0, -self.height)

        # Clear bg
        ctx.clearRect(0, 0, self.width, self.height)

        # Determine drawing area
        x0 = 45
        y0 = 35
        width = self.width - x0 - 15
        height = self.height - y0 - 5

        data = data_per_db[self.dbname]
        if len(data) == 0:
            return

        # Get bounding box
        t1 = data[0].time_start
        t2 = data[-1].time_stop
        mi, ma = self._get_min_max()
        if ma <= mi:
            return
        hscale = width / (t2 - t1)
        vscale = height / (ma - mi)

        unix_from_utc_tuple = Date.UTC  # avoid triggering new
        utc = get_hash_info().get("utc", False)
        xticks = {}

        # Prepare x ticks for hours (one hour is the smallest granularity)
        hourly_tick_units = (1, 3600), (2, 7200), (6, 21600)
        min_tick_dist = 60
        for nhours, tick_unit in hourly_tick_units:
            if tick_unit * hscale >= min_tick_dist:
                break
        else:
            tick_unit = 0
        #
        if tick_unit > 0:
            d = Date(t1 * 1000)
            if utc:
                tup = [
                    d.getUTCFullYear(),
                    d.getUTCMonth(),
                    d.getUTCDate(),
                    d.getUTCHours(),
                ]
                tup[-1] = nhours * int(tup[-1] / nhours)
                t = unix_from_utc_tuple(tup[0], tup[1], tup[2], tup[3]) / 1000
            else:
                tup = [d.getFullYear(), d.getMonth(), d.getDate(), d.getHours()]
                tup[-1] = nhours * int(tup[-1] / nhours)
                t = Date(tup[0], tup[1], tup[2], tup[3]).getTime() / 1000
            while t <= t2:
                if t >= t1:
                    d = Date(t * 1000)
                    if utc:
                        xticks[t] = f"{d.getUTCHours():02i}:{d.getUTCMinutes():02i}"
                    else:
                        xticks[t] = f"{d.getHours():02i}:{d.getMinutes():02i}"
                t += tick_unit

        # Prepare x ticks for days/months
        day_tick_units = (2, 1), (2, 2), (2, 5), (1, 1), (1, 2), (1, 3), (0, 365)
        min_tick_dist = 60
        for dindex, nsomething in day_tick_units:
            tick_unit = nsomething * [365 * 86400, 30 * 86400, 86400][dindex]
            if tick_unit * hscale >= min_tick_dist:
                break
        else:
            tick_unit = nsomething = 0
        #
        n_date_ticks = 0
        if nsomething > 0:
            d = Date(t1 * 1000)
            if utc:
                tup = [d.getUTCFullYear(), d.getUTCMonth(), d.getUTCDate()]
                tup[dindex] = nsomething * int(tup[dindex] / nsomething)
                t = unix_from_utc_tuple(tup[0], tup[1], tup[2]) / 1000
            else:
                tup = [d.getFullYear(), d.getMonth(), d.getDate()]
                tup[dindex] = nsomething * int(tup[dindex] / nsomething)
                t = Date(tup[0], tup[1], tup[2]).getTime() / 1000
            while t <= t2:
                if t >= t1:
                    n_date_ticks += 1
                    d = Date(t * 1000)
                    if utc:
                        dd = f"{d.getUTCDate():02i}"
                        mm = f"{d.getUTCMonth()+1:02i}"
                        yy = f"{d.getUTCFullYear()}"
                        xticks[t] = f"{dd}-{mm}-{yy}"
                    else:
                        dd = f"{d.getDate():02i}"
                        mm = f"{d.getMonth()+1:02i}"
                        yy = f"{d.getFullYear()}"
                        xticks[t] = f"{dd}-{mm}-{yy}"
                tup[dindex] += nsomething
                if utc:
                    t = unix_from_utc_tuple(tup[0], tup[1], tup[2]) / 1000
                else:
                    t = Date(tup[0], tup[1], tup[2]).getTime() / 1000
        #
        extra_x_tick = ""
        if n_date_ticks < 2:
            xtickskeys = xticks.keys()
            if len(xtickskeys) > 0 and hscale * (xtickskeys[0] - t1) < 30:
                xticks.pop(xtickskeys[0])
            d = Date(t1 * 1000)
            if utc:
                extra_x_tick = (
                    f"{d.getUTCFullYear()}-{d.getUTCMonth()+1:02i}-{d.getUTCDate():02i}"
                )
            else:
                extra_x_tick = (
                    f"{d.getFullYear()}-{d.getMonth()+1:02i}-{d.getDate():02i}"
                )

        # Prepare y ticks
        yticks = self._get_ticks(vscale, mi, ma, 25)  # text -> value

        # Prepare drawing
        ctx.lineWidth = 1

        # Draw grid lines
        ctx.strokeStyle = "rgba(128, 128, 128, 0.3)"
        ctx.beginPath()
        for v, text in yticks.items():
            y = y0 + (float(v) - mi) * vscale
            ctx.moveTo(x0, y)
            ctx.lineTo(x0 + width, y)
        ctx.stroke()

        # Draw x ticks
        ctx.strokeStyle = text_color
        ctx.fillStyle = text_color
        ctx.textAlign = "center"
        ctx.textBaseline = "top"  # middle
        ctx.beginPath()
        for t, text in xticks.items():
            x = x0 + (float(t) - t1) * hscale
            ctx.moveTo(x, y0)
            ctx.lineTo(x, y0 - 4)
        ctx.stroke()
        for t, text in xticks.items():
            x = x0 + (float(t) - t1) * hscale
            angle = 0  # -0.15 * Math.PI
            x = min(x, x0 + width - 15)
            self._draw_text(ctx, text, x, y0 - 10, angle)
        if extra_x_tick:
            ctx.textAlign = "left"
            ctx.textBaseline = "bottom"
            self._draw_text(ctx, extra_x_tick, 0, 0)

        # Draw y ticks
        ctx.textAlign = "right"
        ctx.textBaseline = "middle"
        ctx.beginPath()
        for v, text in yticks.items():
            y = y0 + (float(v) - mi) * vscale
            ctx.moveTo(x0 - 4, y)
            ctx.lineTo(x0, y)
        ctx.stroke()
        for v, text in yticks.items():
            y = y0 + (float(v) - mi) * vscale
            self._draw_text(ctx, text, x0 - 8, y)

        # Draw axis
        ctx.strokeStyle = text_color
        ctx.beginPath()
        ctx.moveTo(x0, y0)
        ctx.lineTo(x0 + width, y0)
        ctx.moveTo(x0, y0)
        ctx.lineTo(x0, y0 + height)
        ctx.stroke()

        # Draw content
        self._draw_content(ctx, mi, ma, t1, t2, x0, y0, hscale, vscale)

        # Draw local / UTC
        ctx.fillStyle = "rgba(128, 128, 128, 0.5)"
        ctx.textAlign = "right"
        ctx.textBaseline = "bottom"
        self._draw_text(ctx, "UTC" if utc else "Local time", self.width, 0)


class CountPanel(PlotPanel):

    _values_are_integer = True

    clr = 50, 250, 50

    def _get_min_max(self):
        PSCRIPT_OVERLOAD = False  # noqa
        key = self.key
        mi = 0
        ma = -9_999_999
        data = data_per_db[self.dbname]
        for i in range(len(data)):
            aggr = data[i]
            v = aggr[key]
            if v is undefined:
                continue
            ma = max(ma, v)
        return mi, ma

    def _draw_content(self, ctx, mi, ma, t1, t2, x0, y0, hscale, vscale):
        PSCRIPT_OVERLOAD = False  # noqa
        key = self.key
        clr = self.clr
        ctx.fillStyle = f"rgba({clr[0]}, {clr[1]}, {clr[2]}, 0.8)"
        data = data_per_db[self.dbname]
        for i in range(len(data)):
            aggr = data[i]
            v = aggr[key]
            if v is undefined:
                continue
            if aggr.time_start > t2:
                continue
            x = x0 + (aggr.time_start - t1) * hscale
            w = (aggr.time_stop - aggr.time_start) * hscale
            w = max(w - 1, 1)
            ctx.fillRect(x, y0, w, v * vscale)


class DailyCountPanel(CountPanel):

    clr = 220, 250, 0

    def _get_min_max(self):
        PSCRIPT_OVERLOAD = False  # noqa
        key = self.key
        mi = 0
        ma = -9_999_999
        self.daily = daily = []
        prev_day = ""
        data = data_per_db[self.dbname]
        for i in range(len(data)):
            aggr = data[i]
            v = aggr[key]
            if v is undefined:
                continue
            day = aggr.time_key[:10]
            if day != prev_day:
                if len(daily) > 0:
                    ma = max(ma, daily[-1][key])
                new_aggr = {"time_start": aggr.time_start, "time_stop": aggr.time_stop}
                new_aggr[key] = aggr[key]
                daily.append(new_aggr)
                prev_day = day
            else:
                daily[-1][key] += v
                daily[-1].time_stop = aggr.time_stop
        if len(daily) > 0:
            ma = max(ma, daily[-1][key])
        return mi, ma

    def _draw_content(self, ctx, mi, ma, t1, t2, x0, y0, hscale, vscale):
        PSCRIPT_OVERLOAD = False  # noqa
        # Draw daily
        key = self.key
        clr = self.clr
        ctx.fillStyle = f"rgba({clr[0]}, {clr[1]}, {clr[2]}, 0.4)"
        for i in range(len(self.daily)):
            aggr = self.daily[i]
            v = aggr[key]
            if aggr.time_start > t2:
                continue
            x = x0 + (aggr.time_start - t1) * hscale
            w = (aggr.time_stop - aggr.time_start) * hscale
            w = max(w - 1, 1)
            ctx.fillRect(x, y0, w, v * vscale)
        # Draw per unit
        super()._draw_content(ctx, mi, ma, t1, t2, x0, y0, hscale, vscale)


class MonthlyCountPanel(CountPanel):

    clr = 250, 200, 0

    def _get_min_max(self):
        PSCRIPT_OVERLOAD = False  # noqa
        key = self.key
        mi = 0
        ma = -9_999_999
        self.monthly = monthly = []
        prev_month = ""
        data = data_per_db[self.dbname]
        for i in range(len(data)):
            aggr = data[i]
            v = aggr[key]
            if v is undefined:
                continue
            month = aggr.time_key[:7]
            if month != prev_month:
                if len(monthly) > 0:
                    ma = max(ma, monthly[-1][key])
                new_aggr = {"time_start": aggr.time_start, "time_stop": aggr.time_stop}
                new_aggr[key] = aggr[key]
                monthly.append(new_aggr)
                prev_month = month
            else:
                monthly[-1][key] += v
                monthly[-1].time_stop = aggr.time_stop
        if len(monthly) > 0:
            ma = max(ma, monthly[-1][key])
        return mi, ma

    def _draw_content(self, ctx, mi, ma, t1, t2, x0, y0, hscale, vscale):
        PSCRIPT_OVERLOAD = False  # noqa
        # Draw monthly
        key = self.key
        clr = self.clr
        ctx.fillStyle = f"rgba({clr[0]}, {clr[1]}, {clr[2]}, 0.4)"
        for i in range(len(self.monthly)):
            aggr = self.monthly[i]
            v = aggr[key]
            if aggr.time_start > t2:
                continue
            x = x0 + (aggr.time_start - t1) * hscale
            w = (aggr.time_stop - aggr.time_start) * hscale
            w = max(w - 1, 1)
            ctx.fillRect(x, y0, w, v * vscale)
        # Draw per unit
        super()._draw_content(ctx, mi, ma, t1, t2, x0, y0, hscale, vscale)


class NumericalPanel(PlotPanel):

    clr = 0, 220, 250

    def __init__(self, *args):
        super().__init__(*args)

    def _get_min_max(self):
        PSCRIPT_OVERLOAD = False  # noqa
        key = self.key
        mi = +1e20
        ma = -1e20
        data = data_per_db[self.dbname]
        for i in range(len(data)):
            aggr = data[i]
            meas = aggr[key]
            if meas is undefined or meas.n == 0:
                continue
            mi = min(mi, meas.min)
            ma = max(ma, meas.max)
        if ma >= mi:
            mi = min(0.8 * ma, mi)  # Select a good min point
            mi = 0
        if self.unit == "%":
            mi = 0
            ma = max(ma, 100)  # percentages can be larger than 100
        return mi, ma

    def _draw_content(self, ctx, mi, ma, t1, t2, x0, y0, hscale, vscale):
        PSCRIPT_OVERLOAD = False  # noqa
        key = self.key
        clr = self.clr
        ctx.fillStyle = f"rgba({clr[0]}, {clr[1]}, {clr[2]}, 0.2)"
        ctx.strokeStyle = f"rgba({clr[0]}, {clr[1]}, {clr[2]}, 1.0)"
        data = data_per_db[self.dbname]
        mean_points = []
        for i in range(len(data)):
            aggr = data[i]
            meas = aggr[key]
            if meas is undefined or meas.n == 0:
                continue
            if aggr.time_start > t2:
                continue

            x = x0 + (aggr.time_start - t1) * hscale
            w = (aggr.time_stop - aggr.time_start) * hscale
            w = max(w, 1)

            # Draw rectangle for min max
            y = y0 + (meas.min - mi) * vscale
            h = (meas.max - meas.min) * vscale
            ctx.fillRect(x, y, w, h)

            # Draw rectangle for std
            mean = meas.mean
            std = (meas.magic / meas.n) ** 0.5  # Welford
            st1 = max(meas.min, mean - std)
            st2 = min(meas.max, mean + std)
            y = y0 + (st1 - mi) * vscale
            h = (st2 - st1) * vscale
            ctx.fillRect(x, y, w, h)

            y = y0 + (mean - mi) * vscale
            mean_points.append((x + 0.3333 * w, y))
            mean_points.append((x + 0.6666 * w, y))

        # Draw mean
        if len(mean_points) > 0:
            ctx.beginPath()
            ctx.moveTo(mean_points[0], mean_points[1])
            for x, y in mean_points:
                ctx.lineTo(x, y)
            ctx.stroke()


window.addEventListener("load", on_init)
window.addEventListener("resize", on_resize)
window.addEventListener("hashchange", on_hash_change)



"""
A user-friendly API to public and private RSA keys.
"""

import base64

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa


class PrivateKey:
    def __init__(self, _key):
        assert isinstance(_key, rsa.RSAPrivateKey)
        self._key = _key

    @classmethod
    def generate(cls, size=2048):
        """Generate a new private key, reprsenting an asymetric RSA key pair."""
        private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=size, backend=default_backend()
        )
        return PrivateKey(private_key)

    @classmethod
    def from_str(cls, s, password):
        """Load a private RSA key from a string."""
        assert isinstance(s, str)
        if "OPENSSH" in s[:100]:
            private_key = serialization.load_ssh_private_key(
                s.encode(),
                password=password.encode() if password else None,
            )
        else:
            private_key = serialization.load_pem_private_key(
                s.replace("_", "\n").encode(),
                password=password.encode() if password else None,
                backend=default_backend(),
            )
        return PrivateKey(private_key)

    def to_str(self, password):
        """Serialize this private RSA key to a string, encrypting it with the
        given password (or not encrypting when password is None.
        The returned string has no newlines.
        """
        if password is None:
            encryption = serialization.NoEncryption()
        else:
            encryption = serialization.BestAvailableEncryption(password.encode())
        return (
            self._key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=encryption,
            )
            .decode()
            .replace("\n", "_")
        )

    def get_id(self):
        """Get a short string that identifies this key. The corresponding
        public key has the same id.
        """
        return self.get_public_key().get_id()

    def get_public_key(self):
        """Get the PublicKey correctponding to this private key."""
        public_key = self._key.public_key()
        return PublicKey(public_key)

    def sign(self, data):
        """Sign the given (bytes) data. Returns the (string) signature."""
        assert isinstance(data, bytes)
        sig = self._key.sign(
            data,
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256(),
        )
        return base64.encodebytes(sig).decode()

    def decrypt(self, encrypted_data):
        """Decrypt (bytes) data that was encrypted with the public key."""
        return self._key.decrypt(
            encrypted_data,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )


class PublicKey:
    def __init__(self, _key):
        assert isinstance(_key, rsa.RSAPublicKey)
        self._key = _key

    @classmethod
    def from_str(cls, s):
        """Load a public RSA key from a string (as produced by to_str())."""
        assert isinstance(s, str)
        assert s.startswith("rsa-pub-")
        # Supporting ssh keys would be nice, but their key type is not always supported
        b = base64.urlsafe_b64decode(s[8:].encode())
        public_key = serialization.load_der_public_key(b, backend=default_backend())
        return PublicKey(public_key)

    def to_str(self):
        """Serialize this public RSA key to a (url-safe) string."""
        x = self._key.public_bytes(
            encoding=serialization.Encoding.DER,  # binary
            format=serialization.PublicFormat.PKCS1,  # raw key
        )
        return "rsa-pub-" + base64.urlsafe_b64encode(x).decode()

    def get_id(self):
        """Get a short string that identifies this key. The corresponding
        private key has the same id.
        """
        return self.to_str().strip()[-10:]

    def verify_data(self, signature, data):
        """Verify that the given signature was the result of signing the
        given data with the corresponding private key. Returns a bool.
        """
        # note: we don't call this verify to avoid confusion with self._key.verify
        # which raises an error upon fail instead of returning a bool.
        assert isinstance(signature, str)
        assert isinstance(data, bytes)
        sig = base64.decodebytes(signature.encode())
        try:
            self._key.verify(
                sig,
                data,
                padding.PSS(
                    mgf=padding.MGF1(hashes.SHA256()),
                    salt_length=padding.PSS.MAX_LENGTH,
                ),
                hashes.SHA256(),
            )
        except InvalidSignature:
            return False
        else:
            return True

    def encrypt(self, data):
        """Encrypt the given (bytes) data, which can subsequently only be
        decrypted with the corresponding private key.
        """
        assert isinstance(data, bytes)
        return self._key.encrypt(
            data,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )



"""
Http API for the mypaas daemon.
"""

import os
import io
import time
import queue
import shutil
import logging
import datetime
import asyncio
import zipfile

from mypaas.server import get_deploy_generator, get_public_key


logger = logging.getLogger("mypaasd")

# Fix encoding
os.environ.setdefault("LC_ALL", "C.UTF-8")

# Keep track of tokens that have been used. These expire after x seconds.
invalid_tokens = queue.deque()  # contains (timestamp, token) tuples

# Keep track of whether a deploy is in progress.
deploy_in_progress = False


# %% Utilities


def authenticate(request):
    """Check if the request comes from someone that has a private key
    that we have have authorized.

    This is done by validating (using the public key) that the token is
    signed correctly. We also make sure that keys can only be used once.
    """

    # Get authentication details
    key_id = request.querydict.get("id", "")  # aka fingerprint
    token = request.querydict.get("token", "")
    signature = request.querydict.get("sig1", "")
    if not token or not signature:
        return None

    # Check the timestamp (first part of the token)
    client_time = int(token.split("-")[0])
    server_time = int(time.time())
    if not (server_time - 5 <= client_time <= server_time):
        return None  # too late (or early)

    # Validate the signature
    public_key = get_public_key(key_id)
    if public_key is None:
        return None
    if not public_key.verify_data(signature, token.encode()):
        return None

    # Ok, but what if someone somehow read the key during its transfer
    # and tries to re-use it?
    for _, invalid_token in invalid_tokens:
        if token == invalid_token:
            return None

    # Clear invalid tokens that have expired and mark this one as invalid
    old = server_time - 10
    while invalid_tokens and invalid_tokens[0][0] < old:
        invalid_tokens.popleft()
    invalid_tokens.append((server_time, token))

    # todo: return string based on "comment" in public key.
    return public_key.get_id()  # fingerprint


def validate_payload(request, payload):
    """Verify that the given payload matches the signature."""
    # Get authentication details
    key_id = request.querydict.get("id", "")  # aka fingerprint
    signature = request.querydict.get("sig2", "")
    if not signature:
        return None

    # Validate the payload
    public_key = get_public_key(key_id)
    if public_key is None:
        return None
    if not public_key.verify_data(signature, payload):
        return None

    return public_key.get_id()


def get_uptime_from_start_time(start_time):
    start_time = start_time.rpartition(".")[0] + "+0000"  # get rid of subsecs
    started = datetime.datetime.strptime(start_time, "%Y-%m-%dT%H:%M:%S%z")
    now = datetime.datetime.now(datetime.timezone.utc)
    nsecs = ori_nsecs = (now - started).seconds
    result = []
    if ori_nsecs >= 86400:
        result.append(f"{int(nsecs / 86400)} days")
        nsecs = nsecs % 86400
    if ori_nsecs >= 3600:
        result.append(f"{int(nsecs / 3600)} hours")
        nsecs = nsecs % 3600
    if ori_nsecs >= 60:
        result.append(f"{int(nsecs / 60)} min")
        nsecs = nsecs % 60
    result.append(f"{int(nsecs)} secs")
    return " ".join(result[:2])


# %% Handlers

MAIN_HTML = """
<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8">
  <title>MyPaas Daemon</title>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="stylesheet" href="/style.css">
</head>
<body>

<h1>MyPaas Daemon</h1>

<p style='max-width: 700px;'>
Hi! This is the MyPaas daemon that handles the deploys. It also measures
the system's CPU, memory and disk usage, as well as the CPU and memory
usage of the other MyPaas services, and sends these measurements to the
stats server.
</p>

</body>
</html>
""".lstrip()


async def main_handler(request):
    """Main entry point"""
    if request.path.startswith("/daemon/"):
        path = request.path[7:]
    else:
        return 404, {}, "404 not found"

    if path == "/":
        return 200, {}, MAIN_HTML
    elif path == "/time":
        return 200, {}, str(int(time.time()))
    elif path == "/push":
        return await push(request)
    else:
        return 404, {}, "404 not found"


async def push(request):
    """Push handler. Authenticate, then return generator."""
    if request.method != "POST":
        return 405, {}, "Invalid request"

    fingerprint = authenticate(request)
    if not fingerprint:
        return 403, {}, "Access denied"

    # Get given file
    payload = await request.get_body(100 * 2 ** 20)  # 100 MiB limit

    # Also validate it
    if not validate_payload(request, payload):
        return 403, {}, "Payload could not be verified."

    # Return generator -> do a deploy while streaming feedback on status
    gen = push_generator(fingerprint, payload)
    return 200, {"content-type": "text/plain"}, gen


async def push_generator(fingerprint, payload):
    """Generator that extracts given zipfile and does the deploy."""
    global deploy_in_progress

    # Make sure that only one push happens at a given time
    if deploy_in_progress:
        yield f"Another deploy is in progress by {deploy_in_progress}. Please wait.\n"
        while deploy_in_progress:
            await asyncio.sleep(1)
            yield "."
    deploy_in_progress = True  # Really, really make sure we set this back to False!

    try:

        logger.warn(f"Deploy invoked by {fingerprint}")  # log
        yield f"Hi! This is the MyPaas server. Let's deploy this!\n"
        yield f"Signature validated with public key (fingerprint {fingerprint}).\n"

        # Extract zipfile
        yield "Extracting ...\n"
        deploy_dir = os.path.expanduser("~/_mypaas/deploy_cache")
        shutil.rmtree(deploy_dir, ignore_errors=True)
        os.makedirs(deploy_dir, exist_ok=True)
        with zipfile.ZipFile(io.BytesIO(payload), "r") as zf:
            zf.extractall(deploy_dir)

        # Deploy
        await asyncio.sleep(0.1)
        for step in get_deploy_generator(deploy_dir):
            yield step + "\n"
            await asyncio.sleep(0.1)

    except Exception as err:
        yield "FAIL: " + str(err)
    finally:
        deploy_in_progress = False



"""
The MyPaas setup script.
"""

import os

try:
    import setuptools  # noqa, analysis:ignore
except ImportError:
    pass  # setuptools allows for "develop", but it's not essential

from distutils.core import setup


def get_version_and_doc(filename):
    ns = dict(__version__="", __doc__="")
    docstatus = 0  # Not started, in progress, done
    for line in open(filename, "rb").read().decode().splitlines():
        if line.startswith("__version__"):
            exec(line.strip(), ns, ns)
        elif line.startswith('"""'):
            if docstatus == 0:
                docstatus = 1
                line = line.lstrip('"')
            elif docstatus == 1:
                docstatus = 2
        if docstatus == 1:
            ns["__doc__"] += line.rstrip() + "\n"
    if not ns["__version__"]:
        raise RuntimeError("Could not find __version__")
    return ns["__version__"], ns["__doc__"]


# Get version and docstring (i.e. long description)
version, doc = get_version_and_doc(
    os.path.join(os.path.dirname(__file__), "mypaas", "__init__.py")
)


client_requires = ["cryptography", "requests", "pyperclip", "toml"]
server_requires = ["uvicorn", "asgineer", "psutil", "fastuaparser", "pscript"]


setup(
    name="mypaas",
    version=version,
    author="Almar Klein",
    author_email="",
    license="2-clause BSD",
    url="https://github.com/almarklein/mypaas",
    keywords="paas, saas, deployment, traefik, docker",
    description="Run your own PaaS using Docker, Traefik, and great monitoring",
    long_description=doc,
    platforms="any",
    python_requires=">=3.6",
    install_requires=client_requires,
    extras_require={"server": server_requires},
    packages=[
        "mypaas",
        "mypaas.utils",
        "mypaas.client",
        "mypaas.server",
        "mypaas.daemon",
        "mypaas.stats",
    ],
    entry_points={"console_scripts": ["mypaas = mypaas.__main__:main"]},
    zip_safe=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



import os
import sys
import time
import getpass
import subprocess

from ._traefik import init_router, restart_router
from ._stats import restart_stats
from ._daemon import restart_daemon
from ._auth import server_key_filename, load_config, save_config
from ..utils import dockercall


def init():
    """Initialize the current machine to be a PaaS. You will be asked
    a few questions, so Traefik and the deploy server can be configured
    correctly.
    """

    print("\n    Hi, welcome to MyPaas!\n")
    time.sleep(1)

    # Load config
    config = load_config()

    print("----- Collecting info ".ljust(80, "-"))
    config["init"] = _collect_info_for_config()

    print("----- Preparing the system ".ljust(80, "-"))

    # Ensure that the root mypaas dir exists
    os.makedirs(os.path.expanduser("~/_mypaas"), exist_ok=True)

    # Write config file
    print("Saving MyPaas configuration")
    config["env"] = config.get("env", {}) or {"EXAMPLE_SECRET": "foobar"}
    save_config(config)

    # Make sure the keyfile is there
    pubkeys_filename = os.path.expanduser(server_key_filename)
    if os.path.isfile(pubkeys_filename):
        print(f"Leaving {pubkeys_filename} (containing public keys) as it is.")
    else:
        print(f"Creating {pubkeys_filename} (for public keys)")
        with open(pubkeys_filename, "wb"):
            pass

    # Make sure the keyfile is there
    pubkeys_filename = os.path.expanduser(server_key_filename)
    if os.path.isfile(pubkeys_filename):
        print(f"Leaving {pubkeys_filename} (containing public keys) as it is.")
    else:
        print(f"Creating {pubkeys_filename} (for public keys)")
        with open(pubkeys_filename, "wb"):
            pass

    # Create Docker network
    print("Creating Docker network 'mypaas-net'")
    dockercall("network", "create", "mypaas-net", fail_ok=True)

    # Traefik also needs to so some setup
    init_router()

    print("----- Done ".ljust(80, "-"))
    print("MyPaas is ready to go.")
    print("Run 'mypaas server restart all' to start your PaaS")
    print()


def _collect_info_for_config():
    """Ask questions and return a config dict."""

    print()
    print("MyPaas needs a domain name that will be used for the API endpoint")
    print("and the dashboard. This can be e.g. admin.mydomain.com")
    print("Note that you must point the DNS record to the IP of this server.")
    print()
    domain = input("    Admin domain for this PaaS: ")
    if domain.lower().startswith(("https://", "http://")):
        domain = domain.split("//", 1)[-1].strip()

    print()
    print("The dashboard will be protected with a username and password.")
    print()
    username = input("    username: ")
    pw = getpass.getpass("    password: ")
    pwhash = _get_password_hash(pw)

    print()
    print("The Traefik router uses Let's Encrypt to get SSL/TSL certificates.")
    print("Let's Encrypt can send an email when something is wrong.")
    print()
    email = input("    Email for Let's Encrypt (optional): ").strip()
    print()

    return {"domain": domain, "web_credentials": f"{username}:{pwhash}", "email": email}


def _get_password_hash(pw):
    p = subprocess.Popen(
        ["openssl", "passwd", "-apr1", "-stdin"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    p.stdin.write(pw.encode() + b"\n")
    hash, _ = p.communicate()
    return hash.decode().strip()


def restart(what):
    """Restart one or all of the MyPaas core processes.
    * all: restart router, stats server, and daemon.
    * daemon: restart the deploy daemon.
    * router: restart the Traefik router, e.g. after editing it's config.
    * stats: restart the stats server.
    """
    what = what.lower()

    if what == "all":
        what = "router stats daemon"
    whats = what.split()
    restarted_some = False

    if "daemon" in whats:
        print()
        print("----- (re)starting MyPaas daemon (as a systemctl service)")
        restart_daemon()
        restarted_some = True

    if "router" in whats:
        print()
        print("----- (re)starting Traefik router (as a Docker container)")
        restart_router()
        restarted_some = True

    if "stats" in whats:
        print()
        print("----- (re)starting stats server (as a Docker container)")
        restart_stats()
        restarted_some = True

    if not restarted_some:
        sys.exit(f"Invalid restart argument: {what}")



"""
The Monitor class that performs the aggregation.
"""

import os
import time
import atexit
import hashlib
import weakref
import logging
import datetime
import threading
from queue import Queue, Empty

from itemdb import ItemDB


logger = logging.getLogger("mypaas_stats")


# Number of seconds for one aggregation unit. Assume about 1000 B per
# record. Then with chunks of 10 minutes, one year will take about 1000
# * 6 * 24 * 365 / 2**20 = ~ 50 MiB per year.
DEFAULT_STEP = 10 * 60  # 10 minutes
TABLE_NAME = "aggregations"


_monitor_instances = weakref.WeakSet()
_write_queue = Queue(10000)
_helper_thread = None


# When Python exits, flush the current record of all monitors
@atexit.register
def _at_exit():  # pragma: no cover
    for m in _monitor_instances:
        m.flush()


def hashit(value):
    """Hash any value by applying md5 to the stringified value.
    Returns an integer.
    """
    if isinstance(value, int):
        return abs(value)
    h = hashlib.md5(str(value).encode())
    return abs(int(h.hexdigest()[:14], 16))  # cut at 7 bytes to fit in int64


# Welford's algorithm allows representing (running) measurements without losing
# much precision, and still being able to calculate the mean and std later.
#
# https://github.com/windelbouwman/lognplot/blob/master/lognplot/src/tsdb/sample.rs
# https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm


def _new_num_agg():
    return {"min": 1e20, "max": 0.0, "n": 0, "mean": 0.0, "magic": 0.0}


def std_from_welford(count, mean, magic):
    """Included for rerefence. The data view must implement this."""
    variance = magic / count  # population variance
    # variance = magic / (count - 1)  # sample variance
    return variance ** 0.5


def merge(aggr1, aggr2):
    """Merge aggr2 into aggr1. The caller is responsible for ensuring that
    aggr1 is a copy to avoid overriding cached data.
    """
    aggr1["time_start"] = min(aggr1["time_start"], aggr2["time_start"])
    aggr1["time_stop"] = max(aggr1["time_stop"], aggr2["time_stop"])
    for key, val2 in aggr2.items():
        if "|" not in key:
            continue
        _, type, *_ = key.split("|")
        if type == "count":
            aggr1[key] = aggr1.get(key, 0) + val2
        elif type == "dcount":
            aggr1[key] = aggr1.get(key, 0) + val2
        elif type == "mcount":
            aggr1[key] = aggr1.get(key, 0) + val2
        elif type == "cat":
            d1 = aggr1.get(key, None)
            if d1 is None:
                aggr1[key] = d1 = {}
            for k, c in val2.items():
                d1[k] = d1.get(k, 0) + c
        elif type == "num":
            d1, d2 = aggr1.get(key, None), aggr2.get(key, None)
            if d1 is None:
                aggr1[key] = d1 = _new_num_agg()
            if d2 is not None:
                d1["min"] = min(d1["min"], d2["min"])
                d1["max"] = max(d1["max"], d2["max"])
                # Gather values
                n1, mean1, magic1 = d1["n"], d1["mean"], d1["magic"]
                n2, mean2, magic2 = d2["n"], d2["mean"], d2["magic"]
                # Welford's merge algorithm
                n = n1 + n2
                mean = (mean1 * n1 + mean2 * n2) / n
                delta = mean2 - mean1
                magic = magic1 + magic2 + (delta * n1) * (delta * n2) / n
                # Store result
                d1["n"], d1["mean"], d1["magic"] = n, mean, magic


class HelperThread(threading.Thread):
    """Thread that helps the store to periodically safe aggregations to disk."""

    def __init__(self):
        super().__init__()
        self.setDaemon(True)

    def run(self):
        t = time.time()
        time1 = t + 1.0
        time10 = t + 10.0

        while True:
            try:
                m, aggr = _write_queue.get(True, 0.1)
                m._write_aggr(aggr)
            except Empty:
                pass
            except Exception:
                time.seep(0.1)

            t = time.time()

            if t > time1:
                time1 = t + 1
                for m in _monitor_instances:
                    try:
                        m._do_each_1_seconds()
                    except Exception:
                        pass

            if t > time10:
                time10 = t + 10
                for m in _monitor_instances:
                    try:
                        m._do_each_10_seconds()
                    except Exception:
                        pass


class Monitor:
    """Object that aggregates data and stores it into a database at
    regular intervals. Different types of aggregations can be used, see
    ``put()`` for details.

    Any given data is processed fast, and aggregated at once. Data is
    collected in-memory, and is written to a database (in a separate
    thread) when the time-block ends, or when Python (cleanly) exits.
    Each timeblock is step seconds (default 10 minutes).

    The monitor writes its aggregations to the database specified by
    filename. Multiple monitors can safely use the same database, also
    on different threads or processes. In a docker container you can
    also use a database file on the host system (via ``--volume``). On
    a web server, for instance, one database could be used for system
    measurements, one for each (container) process, and one for each
    domain endpoint.
    """

    def __init__(self, filename, *, step=DEFAULT_STEP):
        self._step = int(step)
        # Prepare db
        self._filename = filename
        # Locks
        self._tlocal = threading.local()  # per-thread data
        self._lock_current_aggr = threading.RLock()
        # Init current aggregation
        self._current_aggr = self._create_new_aggr()
        self._current_time_stop = self._current_aggr["time_stop"]
        # Keep track of ids for daily counters
        self._daily_ids = {}  # key -> set of ids, gets cleared each day
        self._monthly_ids = {}
        if os.path.isfile(self._filename):
            db = ItemDB(self._filename)
            try:
                db.ensure_table("info", "!key")
                daily_ids_info = db.select_one("info", "key == 'daily_ids'")
                day_key = self._current_aggr["time_key"][:10]
                if daily_ids_info and daily_ids_info["time_key"][:10] == day_key:
                    for key in daily_ids_info:
                        if key not in ("key", "time_key"):
                            self._daily_ids[key] = set(daily_ids_info[key])
                monthly_ids_info = db.select_one("info", "key == 'monthly_ids'")
                month_key = self._current_aggr["time_key"][:7]
                if monthly_ids_info and monthly_ids_info["time_key"][:7] == month_key:
                    for key in monthly_ids_info:
                        if key not in ("key", "time_key"):
                            self._monthly_ids[key] = set(monthly_ids_info[key])
            except Exception as err:
                logger.error(
                    f"Failed to restore daily_ids and monthly_ids from db: {err}"
                )
        # Setup our helper thread
        _monitor_instances.add(self)
        global _helper_thread
        if _helper_thread is None:
            _helper_thread = HelperThread()
            _helper_thread.start()

    def _is_locked_in_this_thread(self):
        tlocal = self._tlocal
        try:
            return tlocal.locked is not None
        except AttributeError:
            return False

    def __enter__(self):
        # if we want to be able to use a monitor in different threads.
        if self._is_locked_in_this_thread():
            raise IOError("Already locked by this thread")
        self._tlocal.locked = time.time()
        self._lock_current_aggr.acquire()
        self._maybe_replace_aggr()  # avoid putting in old time-frame
        return self

    def __exit__(self, type, value, traceback):
        self._lock_current_aggr.release()
        self._tlocal.locked = None

    def flush(self):
        """Flush the current aggregation to disk. Mainly used for testing;
        the monitor automatically flushes to the database when the time
        block ends and when Python exits.
        """
        self._write_aggr(self._next_aggr())

    @property
    def filename(self):
        """The filename of the database that this Monitor writes to."""
        return self._filename

    def _do_each_1_seconds(self):
        """Gets called by the helper thread about each second.
        Only do stuff that takes a very short time here!
        """
        pass

    def _do_each_10_seconds(self):
        """Gets called by the helper thread about each 10 seconds.
        Only do stuff that takes a very short time here!
        """
        # For e.g. the SiteMonitor, if no requests come in, we at least
        # flush it
        self._maybe_replace_aggr()

    def _create_new_aggr(self):
        """Create a fresh aggregation object."""
        # Generate a key representing the time corresponding to
        # the start of the current aggregarion block.
        # Format is "yyyy-mm-dd HH:MM:SS". Time is in UTC.
        step = self._step
        block_time = int(time.time() / step) * step
        time_key = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(block_time))
        # Prepare new aggregation object
        new_aggr = {}
        new_aggr["time_key"] = time_key
        new_aggr["time_start"] = int(time.time())  # actual start time
        new_aggr["time_stop"] = block_time + step  # expected stop time
        return new_aggr

    def _next_aggr(self):
        """Replace the current aggregation object and return the old one."""
        # Swap current with a new aggr
        new_aggr = self._create_new_aggr()
        with self._lock_current_aggr:
            cur_aggr = self._current_aggr
            self._current_aggr = new_aggr
            self._current_time_stop = new_aggr["time_stop"]
        # Actual stop time can be earlier
        cur_aggr["time_stop"] = min(cur_aggr["time_stop"], int(time.time()))
        # Return, the helper thread stores it to disk
        return cur_aggr

    def _maybe_replace_aggr(self):
        """Check whether we should replace the current aggregation
        with a new one.
        """
        if time.time() > self._current_time_stop:
            # Swap out the old aggr and have the helper thread store it
            old_aggr = self._next_aggr()
            _write_queue.put((self, old_aggr))
            # Is this a new day?
            old_day = old_aggr["time_key"][:10]
            new_day = self._current_aggr["time_key"][:10]
            if new_day != old_day:
                self._daily_ids = {}
            # Is this a new month?
            old_month = old_aggr["time_key"][:7]
            new_month = self._current_aggr["time_key"][:7]
            if new_month != old_month:
                self._monthly_ids = {}

    def _write_aggr(self, aggr):
        """Write the given aggr to disk. Used by the helper thread to write
        aggr's that we put on the _write_queue.
        """
        for key in aggr.keys():
            if not key.startswith("time_"):
                break
        else:
            return  # Nothing in here, return now
        try:
            db = ItemDB(self._filename)
            # Write aggegation
            db.ensure_table(TABLE_NAME, "!time_key")
            with db:
                x = db.select_one(TABLE_NAME, "time_key == ?", aggr["time_key"])
                if x is not None:
                    merge(x, aggr)
                    aggr = x
                db.put(TABLE_NAME, aggr)
            # Prepare daily ids info
            daily_ids_info = {}
            for key in self._daily_ids.keys():
                daily_ids_info[key] = list(self._daily_ids[key])
            daily_ids_info["key"] = "daily_ids"
            daily_ids_info["time_key"] = self._current_aggr["time_key"][:10]
            # Prepare montly ids info
            monthly_ids_info = {}
            for key in self._monthly_ids.keys():
                monthly_ids_info[key] = list(self._monthly_ids[key])
            monthly_ids_info["key"] = "monthly_ids"
            monthly_ids_info["time_key"] = self._current_aggr["time_key"][:7]
            # Write
            db.ensure_table("info", "!key")
            with db:
                db.put("info", daily_ids_info)
                db.put("info", monthly_ids_info)
        except Exception as err:
            logger.error("Failed to save aggregations: " + str(err))

    def put(self, key, value=None):
        """Put a value into the aggregation. Can only be used under
        the context of this object. Returns True if the value was accepted.

        The key should be of the form "<name>|<type>|<unit>". The unit
        is optional, recognized units include "iB", "s", "%". These
        are handled apropriately.

        The type can be:

        * count: Simply count occurances. Aggregating is summing. The
          value is added to the count, but can also be omitted (default 1).
        * dcount: Count stuff daily. Aggregating is summing, the sum
          over a day is all that really counts. Values are only accepted if
          the given value (a hashable object) has not been seen this (UTC) day.
        * mcount: Same ast dcount, but per month, e.g. unique site visitors.
        * cat: a categorical value. Aggregating is summing the items.
          The value is a string. If ir contains " - " then the left part is
          considered a group to be used while sorting the values for display.
        * num: a numeric value. Aggregating tracks min, max, mean and std.
        """
        # todo: add wcount
        if not self._is_locked_in_this_thread():
            raise IOError("Can only put() under a context.")
        # Pare input
        parts = key.split("|")
        if len(parts) == 2:
            name, type = parts
            unit = ""
        elif len(parts) == 3:
            name, type, unit = parts
        else:
            raise ValueError(
                f"put() key needs name|type or name|type|unit, not {key!r}"
            )
        # Triage over type
        try:
            if type == "count":
                value = 1 if value is None else int(value)
                self._current_aggr[key] = self._current_aggr.get(key, 0) + value
                return True
            elif type == "dcount":
                if value is not None:
                    value = hashit(value)
                    ids = self._daily_ids.setdefault(key, set())
                    if value not in ids:
                        ids.add(value)
                        self._current_aggr[key] = self._current_aggr.get(key, 0) + 1
                        return True
            elif type == "mcount":
                if value is not None:
                    value = hashit(value)
                    ids = self._monthly_ids.setdefault(key, set())
                    if value not in ids:
                        ids.add(value)
                        self._current_aggr[key] = self._current_aggr.get(key, 0) + 1
                        return True
            elif type == "cat":
                if value is not None:
                    value = str(value)
                    if value:
                        d = self._current_aggr.setdefault(key, {})
                        d[value] = d.get(value, 0) + 1
                        return True
            elif type == "num":
                if value is not None:
                    value = float(value)
                    d = self._current_aggr.get(key, None)
                    if d is None:
                        d = _new_num_agg()
                        self._current_aggr[key] = d
                    d["min"] = min(value, d["min"])
                    d["max"] = max(value, d["max"])
                    n1, mean1, magic1 = d["n"], d["mean"], d["magic"]
                    # -- Native implementation, implementing merge for n = 1.
                    # n = n1 + 1
                    # mean = (mean1 * n1 + value) / n
                    # delta = value - mean1
                    # magic = magic1 + (delta * n1) * delta / n
                    # -- Welford online algorithm. Shortcuts -> higher precision
                    n = n1 + 1
                    mean = mean1 + (value - mean1) / n
                    magic = magic1 + (value - mean1) * (value - mean)
                    # Store
                    d["n"], d["mean"], d["magic"] = n, mean, magic
                    return True
            else:
                raise NameError("Unknown aggregation type")
        except Exception as err:
            logger.error(f"Failed to put {type} aggregation {key}: {err}")

    def get_current_aggr(self):
        """Get (a copy of) the current aggregation record."""
        with self._lock_current_aggr:
            return self._current_aggr.copy()

    def get_aggregations(self, first_day, last_day):
        """Get aggregations between two given days (inclusive).
        If the last day is today, also include the current aggregation.
        """
        assert isinstance(first_day, datetime.date)
        assert isinstance(last_day, datetime.date)
        today = time.gmtime()  # UTC
        today = datetime.date(today.tm_year, today.tm_mon, today.tm_mday)
        one_day = datetime.timedelta(days=1)

        data = []

        db = ItemDB(self.filename)
        try:
            data = db.select(
                TABLE_NAME,
                "time_key >= ? AND time_key < ?",
                first_day.strftime("%Y-%m-%d"),
                (last_day + one_day).strftime("%Y-%m-%d"),
            )
        except KeyError:
            pass  # Invalid table name

        if last_day == today:
            data.append(self.get_current_aggr())

        return data



import sys
import time
import subprocess


def restart_daemon():
    """Restart the mypaas daemon."""
    filename = "/etc/systemd/system/mypaasd.service"
    with open(filename, "bw") as f:
        f.write(service.encode())
    time.sleep(0.1)
    try:
        subprocess.check_call(["systemctl", "daemon-reload"])
        subprocess.check_call(["systemctl", "restart", "mypaasd"])
        subprocess.check_call(["systemctl", "enable", "mypaasd"])
    except subprocess.SubprocessError:
        sys.exit("Could not create mypaas daemon service")


service = """
[Unit]
Description=MyPaas daemon
After=network.target
StartLimitIntervalSec=0

[Service]
Type=simple
Restart=always
User=root
WorkingDirectory=/root
ExecStart=/usr/bin/python3 -m mypaas.daemon
RestartSec=2
Environment=MYPAAS_SERVICE=daemon
Environment=MYPAAS_CONTAINER=daemon

[Install]
WantedBy=multi-user.target
"""

# Note: the MYPAAS_CONTAINER is for the deamon's statsgen to
# discover this process and measure its CPU and ram.
